import { __rest } from 'tslib';
import * as i0 from '@angular/core';
import { Injectable, InjectionToken, inject, ElementRef, NgZone, Inject, Optional, SkipSelf, EventEmitter, Directive, Input, Output, Component, ChangeDetectionStrategy, HostBinding, ViewChild, NgModule, ContentChild, Self, Pipe } from '@angular/core';
import * as i1 from 'rxjs';
import { BehaviorSubject, ReplaySubject, take, filter, map, distinctUntilChanged, share, takeUntil, isObservable, combineLatest, Subject, of, tap, concatMap, scheduled, queueScheduler, withLatestFrom, startWith, noop, Observable, Subscription, asapScheduler, pipe, debounceTime, fromEvent, switchMap, pairwise, merge, forkJoin, defer, catchError, throwError } from 'rxjs';
import * as THREE from 'three';
import * as i2 from '@angular/common';
import { DOCUMENT, CommonModule, NgForOf } from '@angular/common';

class Ref extends BehaviorSubject {
    constructor(value) {
        super((value ? value : null));
    }
    set(value) {
        if (typeof value === 'function') {
            this.next(value(this.value));
        }
        else {
            this.next(value);
        }
    }
}

/**
 * Implementation from ngrx/component-store with slight modification
 */
class NgtComponentStore {
    constructor() {
        // Should be used only in ngOnDestroy.
        this.destroySubject$ = new ReplaySubject(1);
        // Exposed to any extending Store to be used for the teardown.
        this.destroy$ = this.destroySubject$.asObservable();
        this.stateSubject$ = new ReplaySubject(1);
    }
    /** Completes all relevant Observable streams. */
    ngOnDestroy() {
        this.stateSubject$.complete();
        this.destroySubject$.next();
    }
    get(projector) {
        let value;
        this.stateSubject$
            .pipe(take(1), skipUndefined(), map((state) => (projector ? projector(state) : state)), skipUndefined())
            .subscribe((result) => {
            value = result;
        });
        return value;
    }
    /**
     * Patches the state with provided partial state.
     *
     * @param partialStateOrUpdaterFn a partial state or a partial updater
     * function that accepts the state and returns the partial state.
     */
    set(partialStateOrUpdaterFn) {
        const patchedState = typeof partialStateOrUpdaterFn === 'function'
            ? partialStateOrUpdaterFn(this.get())
            : partialStateOrUpdaterFn;
        this.updateState((state, partialState) => (Object.assign(Object.assign({}, state), partialState)))(patchedState);
    }
    select(...args) {
        if (args.length === 0) {
            return this.stateSubject$.pipe(skipUndefined(), distinctUntilChanged(), share({
                connector: () => new ReplaySubject(1),
                resetOnComplete: true,
                resetOnRefCountZero: true,
                resetOnError: true,
            }), takeUntil(this.destroy$));
        }
        // if last item is an observable, then the project is missing
        if (isObservable(args[args.length - 1])) {
            const defaultProjector = (() => ({}));
            args.push(defaultProjector);
        }
        const { observables, projector } = processSelectorArgs(args);
        let observable$;
        // If there are no Observables to combine, then we'll just map the value.
        if (observables.length === 0) {
            observable$ = this.stateSubject$.pipe(skipUndefined(), map(projector));
        }
        else {
            // If there are multiple arguments, then we're aggregating selectors, so we need
            // to take the combineLatest of them before calling the map function.
            observable$ = combineLatest(observables).pipe(map((projectorArgs) => projector(...projectorArgs)));
        }
        return observable$.pipe(skipUndefined(), distinctUntilChanged(), share({
            connector: () => new ReplaySubject(1),
            resetOnComplete: true,
            resetOnRefCountZero: true,
            resetOnError: true,
        }), takeUntil(this.destroy$));
    }
    /**
     * Creates an effect.
     *
     * This effect is subscribed to throughout the lifecycle of the ComponentStore.
     * @param generator A function that takes an origin Observable input and
     *     returns an Observable. The Observable that is returned will be
     *     subscribed to for the life of the component.
     * @return A function that, when called, will trigger the origin Observable.
     */
    effect(generator) {
        const origin$ = new Subject();
        generator(origin$)
            // tied to the lifecycle 👇 of ComponentStore
            .pipe(takeUntil(this.destroy$))
            .subscribe();
        return ((observableOrValue) => {
            const observable$ = isObservable(observableOrValue)
                ? observableOrValue
                : of(observableOrValue);
            return observable$
                .pipe(takeUntil(this.destroy$))
                .subscribe((value) => {
                // any new 👇 value is pushed into a stream
                origin$.next(value);
            });
        });
    }
    onCanvasReady(ready$, callback, useEffect = false) {
        const operator = useEffect ? tapEffect : tap;
        return this.effect(operator(() => callback()))(ready$);
    }
    subscribeTo(obs) {
        return this.effect(($) => $)(obs);
    }
    updateState(updaterFn) {
        return ((observableOrValue) => {
            const observable$ = isObservable(observableOrValue)
                ? observableOrValue
                : of(observableOrValue);
            return observable$
                .pipe(concatMap((value) => scheduled([value], queueScheduler).pipe(withLatestFrom(this.stateSubject$.pipe(startWith({}))))), takeUntil(this.destroy$))
                .subscribe({
                next: ([value, currentState]) => {
                    this.stateSubject$.next(updaterFn((currentState || {}), value));
                },
                error: (error) => {
                    this.stateSubject$.error(error);
                },
            });
        });
    }
}
NgtComponentStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtComponentStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgtComponentStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtComponentStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtComponentStore, decorators: [{
            type: Injectable
        }] });
/**
 * An extended `tap` operator that accepts an `effectFn` which:
 * - runs on every `next` notification from `source$`
 * - can optionally return a `cleanUp` function that
 * invokes from the 2nd `next` notification onward and on `unsubscribe` (destroyed)
 *
 *
 * @example
 * ```typescript
 * source$.pipe(
 *  tapEffect((sourceValue) = {
 *    const cb = () => {
 *      doStuff(sourceValue);
 *    };
 *    addListener('event', cb);
 *
 *    return () => {
 *      removeListener('event', cb);
 *    }
 *  })
 * )
 * ```
 */
function tapEffect(effectFn) {
    let cleanupFn = noop;
    let firstRun = false;
    let prev = undefined;
    const teardown = (error) => {
        return () => {
            if (cleanupFn) {
                cleanupFn({ prev, complete: true, error });
            }
        };
    };
    return tap({
        next: (value) => {
            if (cleanupFn && firstRun) {
                cleanupFn({ prev, complete: false, error: false });
            }
            const cleanUpOrVoid = effectFn(value);
            if (cleanUpOrVoid) {
                cleanupFn = cleanUpOrVoid;
            }
            prev = value;
            if (!firstRun) {
                firstRun = true;
            }
        },
        complete: teardown(false),
        unsubscribe: teardown(false),
        error: teardown(true),
    });
}
function debounceSync() {
    return (source) => new Observable((observer) => {
        let actionSubscription;
        let actionValue;
        const rootSubscription = new Subscription();
        rootSubscription.add(source.subscribe({
            complete: () => {
                if (actionSubscription) {
                    observer.next(actionValue);
                }
                observer.complete();
            },
            error: (error) => {
                observer.error(error);
            },
            next: (value) => {
                actionValue = value;
                if (!actionSubscription) {
                    actionSubscription = asapScheduler.schedule(() => {
                        observer.next(actionValue);
                        actionSubscription = undefined;
                    });
                    rootSubscription.add(actionSubscription);
                }
            },
        }));
        return rootSubscription;
    });
}
function skipUndefined() {
    return filter((value) => value !== undefined);
}
function startWithUndefined() {
    return startWith(undefined);
}
function processSelectorArgs(args) {
    const selectorArgs = Array.from(args);
    // Last argument is either projector or config
    const projector = selectorArgs.pop();
    // The Observables to combine, if there are any.
    const observables = selectorArgs;
    return {
        observables,
        projector,
    };
}

const NGT_INSTANCE_REF = new InjectionToken('NgtInstance ref');
const NGT_INSTANCE_HOST_REF = new InjectionToken('NgtInstance ref of the host of the wrapper');
const NGT_OBJECT_REF = new InjectionToken('NgtObject ref');
const NGT_OBJECT_HOST_REF = new InjectionToken('NgtObject ref of the host of the wrapper');
const NGT_CAMERA_REF = new InjectionToken('Ref for root Camera');
const NGT_SCENE_REF = new InjectionToken('Ref for root Scene');
const NGT_COMMON_ATTRIBUTE_REF = new InjectionToken('NgtCommonAttribute ref');
const NGT_COMMON_AUDIO_REF = new InjectionToken('NgtCommonAudio ref');
const NGT_COMMON_CAMERA_REF = new InjectionToken('NgtCommonCamera ref');
const NGT_COMMON_CURVE_REF = new InjectionToken('NgtCommonCurve ref');
const NGT_COMMON_GEOMETRY_REF = new InjectionToken('NgtCommonGeometry ref');
const NGT_COMMON_HELPER_REF = new InjectionToken('NgtCommonHelper ref');
const NGT_COMMON_LIGHT_REF = new InjectionToken('NgtCommonLight ref');
const NGT_COMMON_LINE_REF = new InjectionToken('NgtCommonLine ref');
const NGT_COMMON_MATERIAL_REF = new InjectionToken('NgtCommonMaterial ref');
const NGT_MATERIAL_GEOMETRY_OBJECT_REF = new InjectionToken('NgtMaterialGeometry ref');
const NGT_COMMON_MESH_REF = new InjectionToken('NgtCommonMesh ref');
const NGT_COMMON_OBJECT_HELPER_REF = new InjectionToken('NgtCommonObjectHelper ref');
const NGT_COMMON_SPRITE_REF = new InjectionToken('NgtCommonSprite ref');
const NGT_COMMON_TEXTURE_REF = new InjectionToken('NgtCommonTexture ref');
const NGT_HOST_SKINNED_MESH_REF = new InjectionToken('NgtSkinnedMesh ref as parent host');
const NGT_HOST_BONE_REF = new InjectionToken('NgtBone ref as parent host');
const NGT_HOST_SKELETON_REF = new InjectionToken('NgtSkeleton ref as parent host');

function checkNeedsUpdate(value) {
    if (value !== null &&
        typeof value === 'object' &&
        'needsUpdate' in value) {
        value['needsUpdate'] = true;
    }
}

const is = {
    obj: (a) => a === Object(a) && !is.arr(a) && typeof a !== 'function',
    material: (a) => !!a && a['isMaterial'],
    geometry: (a) => !!a && a['isBufferGeometry'],
    orthographic: (a) => !!a && a.isOrthographicCamera,
    perspective: (a) => !!a && a.isPerspectiveCamera,
    camera: (a) => !!a && a.isCamera,
    glRenderer: (a) => !!a && a instanceof THREE.WebGLRenderer,
    scene: (a) => !!a && a.isScene,
    object3d: (a) => !!a && a.isObject3D,
    instance: (a) => !!a && !!a['__ngt__'],
    ref: (a) => !!a && a instanceof Ref,
    vector3: (a) => !!a && a.isVector3,
    color: (a) => !!a && a.isColor,
    fun: (a) => typeof a === 'function',
    str: (a) => typeof a === 'string',
    num: (a) => typeof a === 'number',
    boo: (a) => typeof a === 'boolean',
    und: (a) => a === void 0,
    arr: (a) => Array.isArray(a),
    equ(a, b, { arrays = 'shallow', objects = 'reference', strict = true, } = {}) {
        // Wrong type or one of the two undefined, doesn't match
        if (typeof a !== typeof b || !!a !== !!b)
            return false;
        // Atomic, just compare a against b
        if (is.str(a) || is.num(a))
            return a === b;
        const isObj = is.obj(a);
        if (isObj && objects === 'reference')
            return a === b;
        const isArr = is.arr(a);
        if (isArr && arrays === 'reference')
            return a === b;
        // Array or Object, shallow compare first to see if it's a match
        if ((isArr || isObj) && a === b)
            return true;
        // Last resort, go through keys
        let i;
        for (i in a)
            if (!(i in b))
                return false;
        for (i in strict ? b : a)
            if (a[i] !== b[i])
                return false;
        if (is.und(i)) {
            if (isArr && a.length === 0 && b.length === 0)
                return true;
            if (isObj &&
                Object.keys(a).length === 0 &&
                Object.keys(b).length === 0)
                return true;
            if (a !== b)
                return false;
        }
        return true;
    },
};

function applyProps(instance, props) {
    var _a;
    // props is empty
    if (!Object.keys(props).length)
        return;
    const instanceInternal = instance.__ngt__ || {};
    const root = instanceInternal.root;
    const rootState = (_a = root === null || root === void 0 ? void 0 : root()) !== null && _a !== void 0 ? _a : {};
    if ('__ngt__' in props) {
        instance.__ngt__ = props
            .__ngt__;
    }
    /** TODO: check if this is still needed */
    // if (
    //     (instance as UnknownRecord)['set'] != null &&
    //     typeof (instance as UnknownRecord)['set'] === 'function' &&
    //     !(instance instanceof THREE.Raycaster)
    // ) {
    //     try {
    //         ((instance as UnknownRecord)['set'] as Function)(props);
    //     } catch (e) {
    //         console.info(
    //             `Swallowing erroneous "set" invoked on ${instance.constructor.name} as non fatal: ${e}`
    //         );
    //     }
    // }
    for (const entry of Object.entries(props)) {
        const key = entry[0];
        let prop = entry[1];
        // raycast is null or undefined. we'll skip
        if (key === 'raycast' && prop == undefined)
            continue;
        if (is.ref(prop)) {
            prop = prop.value;
        }
        const currentInstance = instance;
        const target = currentInstance[key];
        if (target &&
            target['set'] &&
            (target['copy'] || target instanceof THREE.Layers)) {
            // If value is an array
            if (is.arr(prop)) {
                if (target['fromArray'])
                    target['fromArray'](prop);
                else
                    target['set'](...prop);
            }
            else if (target['copy'] &&
                prop &&
                prop.constructor &&
                target.constructor.name ===
                    prop.constructor.name) {
                target['copy'](prop);
            } // If nothing else fits, just set the single value, ignore undefined
            // https://github.com/pmndrs/react-three-fiber/issues/274
            else if (prop !== undefined) {
                const isColor = target instanceof THREE.Color;
                // Allow setting array scalars
                if (!isColor && target['setScalar']) {
                    target['setScalar'](prop);
                }
                // Layers have no copy function, we must therefore copy the mask property
                else if (target instanceof THREE.Layers &&
                    prop instanceof THREE.Layers) {
                    target.mask = prop.mask;
                }
                // Otherwise just set ...
                else {
                    target['set'](prop);
                }
                // For versions of three which don't support THREE.ColorManagement,
                // Auto-convert sRGB colors
                // https://github.com/pmndrs/react-three-fiber/issues/344
                const supportsColorManagement = THREE.ColorManagement;
                if (!supportsColorManagement && !rootState.linear && isColor) {
                    target.convertSRGBToLinear();
                }
            }
        }
        else {
            currentInstance[key] = prop;
            // Auto-convert sRGB textures, for now ...
            // https://github.com/pmndrs/react-three-fiber/issues/344
            if (!rootState.linear &&
                currentInstance[key] instanceof THREE.Texture) {
                currentInstance[key].encoding = THREE.sRGBEncoding;
            }
        }
        checkNeedsUpdate(prop);
        checkNeedsUpdate(target);
    }
}

function coerceBooleanProperty(value) {
    return value != null && `${value}` !== 'false';
}
function coerceNumberProperty(value, fallbackValue = 0) {
    return _isNumberValue(value) ? Number(value) : fallbackValue;
}
/**
 * Whether the provided value is considered a number.
 */
function _isNumberValue(value) {
    // parseFloat(value) handles most of the cases we're interested in (it treats null, empty string,
    // and other non-number values as NaN, where Number just uses 0) but it considers the string
    // '123hello' to be a valid number. Therefore we also check if Number(value) is NaN.
    return !isNaN(parseFloat(value)) && !isNaN(Number(value));
}

const getInstanceRootState = (obj) => { var _a; return (_a = obj.__ngt__) === null || _a === void 0 ? void 0 : _a.root(); };
function prepare(instance, root, parentInstance, previousInstance, isPrimitive = false) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o;
    const parent = parentInstance
        ? parentInstance
        : previousInstance
            ? (_b = (_a = previousInstance.value) === null || _a === void 0 ? void 0 : _a.__ngt__) === null || _b === void 0 ? void 0 : _b.parent
            : undefined;
    if (is.scene(instance)) {
        applyProps(instance, {
            dispose: null,
        });
    }
    return Object.assign(instance, {
        __ngt__: {
            root,
            primitive: !isPrimitive
                ? (_d = (_c = previousInstance === null || previousInstance === void 0 ? void 0 : previousInstance.value) === null || _c === void 0 ? void 0 : _c.__ngt__) === null || _d === void 0 ? void 0 : _d.primitive
                : isPrimitive,
            eventCount: (_g = (_f = (_e = previousInstance === null || previousInstance === void 0 ? void 0 : previousInstance.value) === null || _e === void 0 ? void 0 : _e.__ngt__) === null || _f === void 0 ? void 0 : _f.eventCount) !== null && _g !== void 0 ? _g : 0,
            handlers: (_k = (_j = (_h = previousInstance === null || previousInstance === void 0 ? void 0 : previousInstance.value) === null || _h === void 0 ? void 0 : _h.__ngt__) === null || _j === void 0 ? void 0 : _j.handlers) !== null && _k !== void 0 ? _k : {},
            objects: (_o = (_m = (_l = previousInstance === null || previousInstance === void 0 ? void 0 : previousInstance.value) === null || _l === void 0 ? void 0 : _l.__ngt__) === null || _m === void 0 ? void 0 : _m.objects) !== null && _o !== void 0 ? _o : new Ref([]),
            parent: parent ? (parent === instance ? null : parent) : null,
        },
    });
}

function makeVector2(input) {
    if (!input)
        return new THREE.Vector2();
    if (input instanceof THREE.Vector2) {
        return input;
    }
    if (is.arr(input)) {
        return new THREE.Vector2(...input);
    }
    return new THREE.Vector2(input, input);
}
function makeVector3(input) {
    if (!input)
        return new THREE.Vector3();
    if (input instanceof THREE.Vector3) {
        return input;
    }
    if (is.arr(input)) {
        return new THREE.Vector3(...input);
    }
    return new THREE.Vector3(input, input, input);
}
function makeVector4(input) {
    if (!input)
        return new THREE.Vector4();
    if (input instanceof THREE.Vector4) {
        return input;
    }
    if (is.arr(input)) {
        return new THREE.Vector4(...input);
    }
    return new THREE.Vector4(input, input, input, input);
}
function make(type, input) {
    if (!input)
        return new type();
    if (input instanceof type) {
        return input;
    }
    return new type(...input);
}
function makeColor(color) {
    if (!color)
        return new THREE.Color();
    if (color instanceof THREE.Color) {
        return color;
    }
    if (is.arr(color)) {
        return new THREE.Color(...color);
    }
    return new THREE.Color(color);
}
function makeDpr(dpr) {
    return is.arr(dpr)
        ? Math.min(Math.max(dpr[0], window.devicePixelRatio), dpr[1])
        : dpr;
}
const idCache = {};
/**
 * Generate a random Id or an Event's uuid
 *
 * @param {NgtIntersection} event
 *
 * @internal
 * @private
 */
function makeId(event) {
    if (event) {
        return ((event.eventObject || event.object).uuid +
            '/' +
            event.index +
            event.instanceId);
    }
    const newId = THREE.MathUtils.generateUUID();
    // ensure not already used
    if (!idCache[newId]) {
        idCache[newId] = true;
        return newId;
    }
    return makeId();
}

/**
 * Release pointer captures.
 * This is called by releasePointerCapture in the API, and when an object is removed.
 */
function releaseInternalPointerCapture(capturedMap, obj, captures, pointerId) {
    const captureData = captures.get(obj);
    if (captureData) {
        captures.delete(obj);
        // If this was the last capturing object for this pointer
        if (captures.size === 0) {
            capturedMap.delete(pointerId);
            captureData.target.releasePointerCapture(pointerId);
        }
    }
}
function removeInteractivity(state, object) {
    const internal = state().internal;
    // Removes every trace of an object from the data store
    internal.interaction = internal.interaction.filter((o) => o !== object);
    internal.initialHits = internal.initialHits.filter((o) => o !== object);
    internal.hovered.forEach((value, key) => {
        if (value.eventObject === object || value.object === object) {
            // Clear out intersects, they are outdated by now
            internal.hovered.delete(key);
        }
    });
    internal.capturedMap.forEach((captures, pointerId) => {
        releaseInternalPointerCapture(internal.capturedMap, object, captures, pointerId);
    });
}
function createEvents(state) {
    const temp = new THREE.Vector3();
    /** Calculates delta */
    function calculateDistance(event) {
        const internal = state().internal;
        const dx = event.offsetX - internal.initialClick[0];
        const dy = event.offsetY - internal.initialClick[1];
        return Math.round(Math.sqrt(dx * dx + dy * dy));
    }
    /** Returns true if an instance has a valid pointer-event registered, this excludes scroll, clicks etc */
    function filterPointerEvents(objects) {
        return objects.filter((obj) => ['move', 'over', 'enter', 'out', 'leave'].some((name) => { var _a, _b; return (_b = (_a = obj.__ngt__) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b[('pointer' + name)]; }));
    }
    function intersect(event, filter) {
        var _a, _b, _c;
        const rootState = state();
        const duplicates = new Set();
        const intersections = [];
        // Allow callers to eliminate event objects
        const eventsObjects = filter
            ? filter(rootState.internal.interaction)
            : rootState.internal.interaction;
        // Reset all raycaster cameras to undefined
        eventsObjects.forEach((obj) => {
            const state = getInstanceRootState(obj);
            if (state) {
                state.raycaster.camera = undefined;
            }
        });
        if (!rootState.previousState) {
            // Make sure root-level pointer and ray are set up
            (_b = (_a = rootState.events).compute) === null || _b === void 0 ? void 0 : _b.call(_a, event, rootState);
        }
        // Collect events
        let hits = eventsObjects
            // Intersect objects
            // @ts-ignore
            .flatMap((obj) => {
            var _a, _b;
            const state = getInstanceRootState(obj);
            // Skip event handling when noEvents is set, or when the raycasters camera is null
            if (!state ||
                !state.events.enabled ||
                state.raycaster.camera === null)
                return [];
            // When the camera is undefined we have to call the event layers update function
            if (state.raycaster.camera === undefined) {
                (_b = (_a = state.events).compute) === null || _b === void 0 ? void 0 : _b.call(_a, event, state, state.previousState);
                // If the camera is still undefined we have to skip this layer entirely
                if (state.raycaster.camera === undefined)
                    state.raycaster.camera = null;
            }
            // Intersect object by object
            return state.raycaster.camera
                ? state.raycaster.intersectObject(obj, true)
                : [];
        })
            // Sort by event priority and distance
            .sort((a, b) => {
            const aState = getInstanceRootState(a.object);
            const bState = getInstanceRootState(b.object);
            if (!aState || !bState)
                return 0;
            return (bState.events.priority - aState.events.priority ||
                a.distance - b.distance);
        })
            // Filter out duplicates
            .filter((item) => {
            const id = makeId(item);
            if (duplicates.has(id))
                return false;
            duplicates.add(id);
            return true;
        });
        // https://github.com/mrdoob/three.js/issues/16031
        // Allow custom userland intersect sort order, this likely only makes sense on the root filter
        if (rootState.events.filter)
            hits = rootState.events.filter(hits, rootState);
        // Bubble up the events, find the event source (eventObject)
        for (const hit of hits) {
            let eventObject = hit.object;
            // Bubble event up
            while (eventObject) {
                if ((_c = eventObject.__ngt__) === null || _c === void 0 ? void 0 : _c.eventCount)
                    intersections.push(Object.assign(Object.assign({}, hit), { eventObject }));
                eventObject = eventObject.parent;
            }
        }
        // If the interaction is captured, make all capturing targets part of the intersect.
        if ('pointerId' in event &&
            rootState.internal.capturedMap.has(event.pointerId)) {
            for (const captureData of rootState.internal.capturedMap
                .get(event.pointerId)
                .values()) {
                intersections.push(captureData.intersection);
            }
        }
        return intersections;
    }
    /**  Handles intersections by forwarding them to handlers */
    function handleIntersects(intersections, event, delta, callback) {
        const { raycaster, pointer, camera, internal } = state();
        // If anything has been found, forward it to the event listeners
        if (intersections.length) {
            const unprojectedPoint = temp
                .set(pointer.x, pointer.y, 0)
                .unproject(camera);
            const localState = { stopped: false };
            for (const hit of intersections) {
                const hasPointerCapture = (id) => { var _a, _b; return (_b = (_a = internal.capturedMap.get(id)) === null || _a === void 0 ? void 0 : _a.has(hit.eventObject)) !== null && _b !== void 0 ? _b : false; };
                const setPointerCapture = (id) => {
                    const captureData = {
                        intersection: hit,
                        target: event.target,
                    };
                    if (internal.capturedMap.has(id)) {
                        // if the pointerId was previously captured, we add the hit to the
                        // event capturedMap.
                        internal.capturedMap
                            .get(id)
                            .set(hit.eventObject, captureData);
                    }
                    else {
                        // if the pointerId was not previously captured, we create a map
                        // containing the hitObject, and the hit. hitObject is used for
                        // faster access.
                        internal.capturedMap.set(id, new Map([[hit.eventObject, captureData]]));
                    }
                    // Call the original event now
                    event.target.setPointerCapture(id);
                };
                const releasePointerCapture = (id) => {
                    const captures = internal.capturedMap.get(id);
                    if (captures) {
                        releaseInternalPointerCapture(internal.capturedMap, hit.eventObject, captures, id);
                    }
                };
                // Add native event props
                const extractEventProps = {};
                // This iterates over the event's properties including the inherited ones. Native PointerEvents have most of their props as getters which are inherited, but polyfilled PointerEvents have them all as their own properties (i.e. not inherited). We can't use Object.keys() or Object.entries() as they only return "own" properties; nor Object.getPrototypeOf(event) as that *doesn't* return "own" properties, only inherited ones.
                for (const prop in event) {
                    const property = event[prop];
                    // Only copy over atomics, leave functions alone as these should be
                    // called as event.nativeEvent.fn()
                    if (typeof property !== 'function')
                        extractEventProps[prop] = property;
                }
                const raycastEvent = Object.assign(Object.assign(Object.assign({}, hit), extractEventProps), { pointer,
                    intersections, stopped: localState.stopped, delta,
                    unprojectedPoint, ray: raycaster.ray, camera: camera, 
                    // Hijack stopPropagation, which just sets a flag
                    stopPropagation: () => {
                        // https://github.com/pmndrs/react-three-fiber/issues/596
                        // Events are not allowed to stop propagation if the pointer has been captured
                        const capturesForPointer = 'pointerId' in event &&
                            internal.capturedMap.get(event.pointerId);
                        // We only authorize stopPropagation...
                        if (
                        // ...if this pointer hasn't been captured
                        !capturesForPointer ||
                            // ... or if the hit object is capturing the pointer
                            capturesForPointer.has(hit.eventObject)) {
                            raycastEvent.stopped = localState.stopped = true;
                            // Propagation is stopped, remove all other hover records
                            // An event handler is only allowed to flush other handlers if it is hovered itself
                            if (internal.hovered.size &&
                                Array.from(internal.hovered.values()).find((i) => i.eventObject === hit.eventObject)) {
                                // Objects cannot flush out higher up objects that have already caught the event
                                const higher = intersections.slice(0, intersections.indexOf(hit));
                                cancelPointer([...higher, hit]);
                            }
                        }
                    }, 
                    // there should be a distinction between target and currentTarget
                    target: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, currentTarget: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, nativeEvent: event });
                // Call subscribers
                callback(raycastEvent);
                // Event bubbling may be interrupted by stopPropagation
                if (localState.stopped === true)
                    break;
            }
        }
        return intersections;
    }
    function cancelPointer(hits) {
        const internal = state().internal;
        Array.from(internal.hovered.values()).forEach((hoveredObj) => {
            var _a, _b;
            // When no objects were hit or the the hovered object wasn't found underneath the cursor
            // we call onPointerOut and delete the object from the hovered-elements map
            if (!hits.length ||
                !hits.find((hit) => hit.object === hoveredObj.object &&
                    hit.index === hoveredObj.index &&
                    hit.instanceId === hoveredObj.instanceId)) {
                const eventObject = hoveredObj.eventObject;
                const instance = eventObject
                    .__ngt__;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                internal.hovered.delete(makeId(hoveredObj));
                if (instance === null || instance === void 0 ? void 0 : instance.eventCount) {
                    // Clear out intersects, they are outdated by now
                    const data = Object.assign(Object.assign({}, hoveredObj), { intersections: hits || [] });
                    (_a = handlers.pointerout) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                    (_b = handlers.pointerleave) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                }
            }
        });
    }
    const handlePointer = (name) => {
        // Deal with cancelation
        switch (name) {
            case 'pointerleave':
            case 'pointercancel':
                return () => cancelPointer([]);
            case 'lostpointercapture':
                return (event) => {
                    const internal = state().internal;
                    if ('pointerId' in event &&
                        !internal.capturedMap.has(event.pointerId)) {
                        // If the object event interface had onLostPointerCapture, we'd call it here on every
                        // object that's getting removed.
                        internal.capturedMap.delete(event.pointerId);
                        cancelPointer([]);
                    }
                };
        }
        // Any other pointer goes here ...
        return (event) => {
            const { pointerMissed, internal } = state();
            //prepareRay(event)
            internal.lastEvent = event;
            // Get fresh intersects
            const isPointerMove = name === 'pointermove';
            const isClickEvent = name === 'click' ||
                name === 'contextmenu' ||
                name === 'dblclick';
            const filter = isPointerMove ? filterPointerEvents : undefined;
            //const hits = patchIntersects(intersect(filter), event)
            const hits = intersect(event, filter);
            const delta = isClickEvent ? calculateDistance(event) : 0;
            // Save initial coordinates on pointer-down
            if (name === 'pointerdown') {
                internal.initialClick = [event.offsetX, event.offsetY];
                internal.initialHits = hits.map((hit) => hit.eventObject);
            }
            // If a click yields no results, pass it back to the user as a miss
            // Missed events have to come first in order to establish user-land side-effect clean up
            if (isClickEvent && !hits.length) {
                if (delta <= 2) {
                    objectPointerMissed(event, internal.interaction);
                    if (pointerMissed)
                        pointerMissed(event);
                }
            }
            // Take care of unhover
            if (isPointerMove)
                cancelPointer(hits);
            handleIntersects(hits, event, delta, (data) => {
                var _a, _b, _c;
                const eventObject = data.eventObject;
                const instance = eventObject
                    .__ngt__;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                // Check presence of handlers
                if (!(instance === null || instance === void 0 ? void 0 : instance.eventCount))
                    return;
                if (isPointerMove) {
                    // Move event ...
                    if (handlers.pointerover ||
                        handlers.pointerenter ||
                        handlers.pointerout ||
                        handlers.pointerleave) {
                        // When enter or out is present take care of hover-state
                        const id = makeId(data);
                        const hoveredItem = internal.hovered.get(id);
                        if (!hoveredItem) {
                            // If the object wasn't previously hovered, book it and call its handler
                            internal.hovered.set(id, data);
                            (_a = handlers.pointerover) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                            (_b = handlers.pointerenter) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                        }
                        else if (hoveredItem.stopped) {
                            // If the object was previously hovered and stopped, we shouldn't allow other items to proceed
                            data.stopPropagation();
                        }
                    }
                    // Call mouse move
                    (_c = handlers.pointermove) === null || _c === void 0 ? void 0 : _c.call(handlers, data);
                }
                else {
                    // All other events ...
                    const handler = handlers[name];
                    if (handler) {
                        // Forward all events back to their respective handlers with the exception of click events,
                        // which must use the initial target
                        if (!isClickEvent ||
                            internal.initialHits.includes(eventObject)) {
                            // Missed events have to come first
                            objectPointerMissed(event, internal.interaction.filter((object) => !internal.initialHits.includes(object)));
                            // Now call the handler
                            handler(data);
                        }
                    }
                    else {
                        // Trigger onPointerMissed on all elements that have pointer over/out handlers, but not click and weren't hit
                        if (isClickEvent &&
                            internal.initialHits.includes(eventObject)) {
                            objectPointerMissed(event, internal.interaction.filter((object) => !internal.initialHits.includes(object)));
                        }
                    }
                }
            });
        };
    };
    function objectPointerMissed(event, objects) {
        objects.forEach((object) => { var _a, _b, _c; return (_c = (_b = (_a = object.__ngt__) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b.pointermissed) === null || _c === void 0 ? void 0 : _c.call(_b, event); });
    }
    return { handlePointer };
}

function mutate(object, path, value) {
    const { decomposedPath, base } = decomposePath(path);
    if (base === undefined) {
        return;
    }
    // assign an empty object in order to spread object
    assignEmpty(object, base);
    // Determine if there is still layers to traverse
    if (decomposedPath.length <= 1) {
        applyProps(object, { [base]: value });
    }
    else {
        mutate(object[base], decomposedPath.slice(1), value);
    }
}
function decomposePath(path) {
    if (path.length < 1) {
        return { base: '', decomposedPath: [] };
    }
    const decomposedPath = path;
    const base = path[0];
    return { base, decomposedPath };
}
function assignEmpty(obj, base) {
    if (!Object.prototype.hasOwnProperty.call(obj, base) &&
        Reflect &&
        !!Reflect.has &&
        !Reflect.has(obj, base)) {
        obj[base] = {};
    }
}

const NGT_PERFORMANCE_OPTIONS = new InjectionToken('NgtPerformance options', {
    factory: () => ({
        current: 1,
        min: 0.5,
        max: 1,
        debounce: 200,
    }),
});
function providePerformanceOptions(options) {
    return {
        provide: NGT_PERFORMANCE_OPTIONS,
        useValue: Object.assign(Object.assign({}, options), { current: 1, min: 0.5, max: 1, debounce: 200 }),
    };
}

const WINDOW = new InjectionToken('window', {
    factory: () => {
        const { defaultView } = inject(DOCUMENT);
        if (!defaultView) {
            throw `window is not available!`;
        }
        return defaultView;
    },
});

const NGT_RESIZE_BOX_DEFAULT = 'content-box';
const NGT_RESIZE_DEBOUNCE_DEFAULT = 200;
const NGT_RESIZE_OPTIONS = new InjectionToken('ngtResizeObserver Options', {
    factory: () => ({
        box: NGT_RESIZE_BOX_DEFAULT,
        scroll: false,
        offsetSize: false,
        debounce: NGT_RESIZE_DEBOUNCE_DEFAULT,
    }),
});
function provideResizeOptions(resizeOptions) {
    return {
        provide: NGT_RESIZE_OPTIONS,
        useValue: Object.assign(Object.assign({}, resizeOptions), { box: NGT_RESIZE_BOX_DEFAULT, scroll: false, offsetSize: false, debounce: NGT_RESIZE_DEBOUNCE_DEFAULT }),
    };
}
const NGT_RESIZE_OBSERVER_SUPPORT = new InjectionToken('Resize Observer API support', {
    factory: () => {
        const window = inject(WINDOW);
        return ('ResizeObserver' in window &&
            window['ResizeObserver'] != null);
    },
});

class NgtResize extends Observable {
    constructor({ nativeElement }, zone, document, window, isSupport, { box, offsetSize, scroll, debounce }) {
        let observer;
        let lastBounds;
        let lastEntries = [];
        const torndown$ = new Subject();
        const scrollContainers = findScrollContainers(nativeElement);
        // set actual debounce values early, so effects know if they should react accordingly
        const scrollDebounce = debounce
            ? typeof debounce === 'number'
                ? debounce
                : debounce.scroll
            : null;
        const resizeDebounce = debounce
            ? typeof debounce === 'number'
                ? debounce
                : debounce.resize
            : null;
        const debounceAndDestroy = (debounce) => {
            return pipe(debounceTime(debounce !== null && debounce !== void 0 ? debounce : 0), takeUntil(torndown$));
        };
        super((subscriber) => {
            if (!isSupport) {
                subscriber.error('ResizeObserver is not supported in your browser. Please use a polyfill');
                return;
            }
            zone.runOutsideAngular(() => {
                const callback = (entries) => {
                    lastEntries = entries;
                    const { left, top, width, height, bottom, right, x, y } = nativeElement.getBoundingClientRect();
                    const size = {
                        left,
                        top,
                        width,
                        height,
                        bottom,
                        right,
                        x,
                        y,
                    };
                    if (nativeElement instanceof HTMLElement && offsetSize) {
                        size.height = nativeElement.offsetHeight;
                        size.width = nativeElement.offsetWidth;
                    }
                    Object.freeze(size);
                    subscriber.next(Object.assign({ entries, dpr: window.devicePixelRatio }, size));
                    if (!areBoundsEqual(lastBounds || {}, size)) {
                        lastBounds = size;
                    }
                };
                const boundEntriesCallback = () => {
                    callback(lastEntries);
                };
                observer = new ResizeObserver(callback);
                observer.observe(nativeElement, { box });
                if (scroll) {
                    if (scrollContainers) {
                        scrollContainers.forEach((scrollContainer) => {
                            fromEvent(scrollContainer, 'scroll', {
                                capture: true,
                                passive: true,
                            })
                                .pipe(debounceAndDestroy(scrollDebounce))
                                .subscribe(boundEntriesCallback);
                        });
                    }
                    fromEvent(window, 'scroll', {
                        capture: true,
                        passive: true,
                    })
                        .pipe(debounceAndDestroy(scrollDebounce))
                        .subscribe(boundEntriesCallback);
                }
                fromEvent(window, 'resize')
                    .pipe(debounceAndDestroy(resizeDebounce))
                    .subscribe(boundEntriesCallback);
            });
            return () => {
                if (observer) {
                    observer.unobserve(nativeElement);
                    observer.disconnect();
                }
                torndown$.next();
                torndown$.complete();
            };
        });
        return this.pipe(debounceTime(scrollDebounce || 0), share({
            connector: () => new ReplaySubject(1),
            resetOnRefCountZero: true,
            resetOnComplete: true,
        }));
    }
}
NgtResize.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtResize, deps: [{ token: ElementRef }, { token: NgZone }, { token: DOCUMENT }, { token: WINDOW }, { token: NGT_RESIZE_OBSERVER_SUPPORT }, { token: NGT_RESIZE_OPTIONS }], target: i0.ɵɵFactoryTarget.Injectable });
NgtResize.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtResize });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtResize, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: i0.ElementRef, decorators: [{
                        type: Inject,
                        args: [ElementRef]
                    }] }, { type: i0.NgZone, decorators: [{
                        type: Inject,
                        args: [NgZone]
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }, { type: Window, decorators: [{
                        type: Inject,
                        args: [WINDOW]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_RESIZE_OBSERVER_SUPPORT]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_RESIZE_OPTIONS]
                    }] }];
    } });
// Returns a list of scroll offsets
function findScrollContainers(element) {
    const result = [];
    if (!element || element === document.body)
        return result;
    const { overflow, overflowX, overflowY } = window.getComputedStyle(element);
    if ([overflow, overflowX, overflowY].some((prop) => prop === 'auto' || prop === 'scroll'))
        result.push(element);
    return [
        ...result,
        ...findScrollContainers(element.parentElement),
    ];
}
// Checks if element boundaries are equal
const keys = [
    'x',
    'y',
    'top',
    'bottom',
    'left',
    'right',
    'width',
    'height',
];
const areBoundsEqual = (a, b) => keys.every((key) => a[key] === b[key]);

function updateCamera(cameraOptions, camera, size) {
    // leave the userland camera alone
    if (!(is.camera(cameraOptions) || camera.manual)) {
        if (is.orthographic(camera)) {
            camera.left = size.width / -2;
            camera.right = size.width / 2;
            camera.top = size.height / 2;
            camera.bottom = size.height / -2;
        }
        else {
            camera.aspect = size.width / size.height;
        }
        camera.updateProjectionMatrix();
        // https://github.com/pmndrs/react-three-fiber/issues/178
        // Update matrix world since the renderer is a frame late
        camera.updateMatrixWorld();
    }
}
function createDefaultCamera(state) {
    if (state.orthographic) {
        return new THREE.OrthographicCamera(0, 0, 0, 0, 0.1, 1000);
    }
    return new THREE.PerspectiveCamera(75, state.size.width / state.size.height, 0.1, 1000);
}

const DOM_EVENTS = {
    click: false,
    contextmenu: false,
    dblclick: false,
    wheel: false,
    pointerdown: true,
    pointerup: true,
    pointerleave: true,
    pointermove: true,
    pointercancel: true,
    lostpointercapture: true,
};
class NgtStore extends NgtComponentStore {
    constructor({ nativeElement: { clientWidth, clientHeight }, }, performanceOptions, { devicePixelRatio }, parentStore, resizeResult$, zone) {
        super();
        this.parentStore = parentStore;
        this.resizeResult$ = resizeResult$;
        this.zone = zone;
        this.pointer = new THREE.Vector2();
        this.position = new THREE.Vector3();
        this.defaultTarget = new THREE.Vector3();
        this.tempTarget = new THREE.Vector3();
        this.ready$ = this.select((s) => s.ready).pipe(filter((ready) => ready));
        this.camera$ = this.select((s) => s.camera);
        this.scene$ = this.select((s) => s.scene);
        this.gl$ = this.select((s) => s.gl);
        this.raycaster$ = this.select((s) => s.raycaster);
        this.active$ = this.select((s) => s.internal.active);
        this.allConstructed$ = this.select(this.camera$, this.scene$, this.gl$, this.raycaster$, this.active$, (camera, scene, gl, raycaster, active) => !!camera && !!gl && !!scene && !!raycaster && active === true).pipe(map((ready) => ({ ready })));
        this.performanceTimeoutId = undefined;
        this.dimensions$ = this.select(this.select((s) => s.size), this.select((s) => s.viewport), (size, viewport) => ({ size, viewport }));
        this.startLoop = this.effect(tap(({ invalidate }) => {
            invalidate();
        }));
        this.initRenderer = this.effect(tapEffect(({ canvasElement, rootStateMap, advance }) => {
            var _a, _b, _c, _d;
            const state = this.get();
            // Scene
            const scene = prepare(new THREE.Scene(), () => this.get(), ((_a = this.parentStore) === null || _a === void 0 ? void 0 : _a.get((s) => s.sceneRef)) || null);
            applyProps(scene, state.sceneOptions);
            this.get((s) => s.sceneRef).set(scene);
            // Set up renderer (one time only!)
            let gl = state.gl;
            if (!state.gl) {
                gl = createRenderer(state.glOptions, canvasElement);
            }
            this.set({ gl });
            // Set up raycaster (one time only!)
            let raycaster = state.raycaster;
            if (!state.raycaster) {
                raycaster = new THREE.Raycaster();
            }
            // Set raycaster options
            const _e = state.raycasterOptions || {}, { params } = _e, options = __rest(_e, ["params"]);
            applyProps(raycaster, Object.assign(Object.assign({ enabled: true }, options), { params: Object.assign(Object.assign({}, raycaster.params), (params || {})) }));
            // Create default camera (one time only!)
            let camera = state.camera;
            if (!state.camera) {
                const isCamera = is.camera(state.cameraOptions);
                camera = isCamera
                    ? state.cameraOptions
                    : createDefaultCamera(state);
                if (!isCamera) {
                    if (state.cameraOptions) {
                        applyProps(camera, state.cameraOptions);
                    }
                    // Set position.z if position not passed in
                    if (!((_b = state.cameraOptions) === null || _b === void 0 ? void 0 : _b.position)) {
                        camera.position.z = 5;
                    }
                    // Always look at center by default
                    if (!((_c = state.cameraOptions) === null || _c === void 0 ? void 0 : _c.rotation)) {
                        if (state.cameraOptions.lookAt) {
                            const lookAt = state.cameraOptions.lookAt;
                            if (is.arr(lookAt)) {
                                camera.lookAt(lookAt[0], lookAt[1], lookAt[2]);
                            }
                            else {
                                camera.lookAt(lookAt);
                            }
                        }
                        else {
                            camera.lookAt(0, 0, 0);
                        }
                    }
                    // Update projection matrix after applying props
                    camera.updateProjectionMatrix();
                }
                if (!is.instance(camera)) {
                    camera = prepare(camera, () => this.get(), ((_d = this.parentStore) === null || _d === void 0 ? void 0 : _d.get((s) => s.cameraRef)) || null);
                }
                this.get((s) => s.cameraRef).set(camera);
            }
            // Set up XR (one time only!)
            let xr = state.xr;
            if (!state.xr) {
                // Handle frame behavior in WebXR
                const handleXRFrame = (timestamp, frame) => {
                    const state = this.get();
                    if (state.frameloop === 'never')
                        return;
                    advance(timestamp, true, () => state, frame);
                };
                // Toggle render switching on session
                const handleSessionChange = () => {
                    const gl = this.get((s) => s.gl);
                    gl.xr.enabled = gl.xr.isPresenting;
                    // WebXRManager's signature is incorrect.
                    // See: https://github.com/pmndrs/react-three-fiber/pull/2017#discussion_r790134505
                    gl.xr.setAnimationLoop(gl.xr.isPresenting ? handleXRFrame : null);
                };
                // WebXR session manager
                xr = {
                    connect: () => {
                        const gl = this.get((s) => s.gl);
                        gl.xr.addEventListener('sessionstart', handleSessionChange);
                        gl.xr.addEventListener('sessionend', handleSessionChange);
                    },
                    disconnect: () => {
                        const gl = this.get((s) => s.gl);
                        gl.xr.removeEventListener('sessionstart', handleSessionChange);
                        gl.xr.removeEventListener('sessionend', handleSessionChange);
                    },
                };
                // Subscribe to WebXR session events
                if (gl.xr)
                    xr.connect();
            }
            // Set shadowmap
            if (gl.shadowMap) {
                if (state.shadows) {
                    gl.shadowMap.enabled = true;
                    if (typeof state.shadows === 'object') {
                        Object.assign(gl.shadowMap, state.shadows);
                    }
                    else {
                        gl.shadowMap.type = THREE.PCFSoftShadowMap;
                    }
                    gl.shadowMap.needsUpdate = true;
                }
            }
            // Set color management
            if (THREE.ColorManagement) {
                THREE.ColorManagement.legacyMode = state.legacy;
            }
            const outputEncoding = state.linear
                ? THREE.LinearEncoding
                : THREE.sRGBEncoding;
            const toneMapping = state.flat
                ? THREE.NoToneMapping
                : THREE.ACESFilmicToneMapping;
            if (gl.outputEncoding !== outputEncoding) {
                gl.outputEncoding = outputEncoding;
            }
            if (gl.toneMapping !== toneMapping) {
                gl.toneMapping = toneMapping;
            }
            gl.setClearAlpha(0);
            gl.setPixelRatio(makeDpr(state.viewport.dpr));
            gl.setSize(state.size.width, state.size.height);
            if (typeof state.glOptions === 'object' &&
                typeof state.glOptions !== 'function' &&
                !is.glRenderer(state.glOptions)) {
                applyProps(gl, state.glOptions);
            }
            this.set({ gl, camera, scene, raycaster });
            return () => {
                const { gl, xr, cameraRef, sceneRef, events } = this.get();
                if (gl) {
                    if (events.disconnect) {
                        events.disconnect();
                    }
                    gl.renderLists.dispose();
                    gl.forceContextLoss();
                    if (gl.xr && gl.xr.enabled) {
                        gl.xr.setAnimationLoop(null);
                        xr.disconnect();
                    }
                    cameraRef.complete();
                    sceneRef.complete();
                    this.set((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { active: false }) })));
                    rootStateMap.delete(canvasElement);
                }
            };
        }));
        this.resize = this.effect(tap(({ width, height, dpr }) => {
            this.set(({ viewport, camera }) => {
                const size = { width, height };
                return {
                    size,
                    viewport: Object.assign(Object.assign(Object.assign({}, viewport), viewport.getCurrentViewport(camera, this.defaultTarget, size)), { dpr: makeDpr(dpr) }),
                };
            });
        }));
        this.updateDimensions = this.effect(tap(({ size, viewport }) => {
            const { camera, gl, ready, cameraOptions } = this.get();
            if (ready) {
                updateCamera(cameraOptions, camera, size);
                // update renderer
                gl.setPixelRatio(viewport.dpr);
                gl.setSize(size.width, size.height);
            }
        }));
        this.registerBeforeRenderEffect = this.effect(tapEffect((_a) => {
            var { uuid } = _a, record = __rest(_a, ["uuid"]);
            if (uuid) {
                this.set((state) => ({
                    internal: Object.assign(Object.assign({}, state.internal), { animations: new Map(state.internal.animations).set(uuid, record), priority: state.internal.priority +
                            ((record.priority || 0) > 0 ? 1 : 0) }),
                }));
            }
            return ({ prev: { uuid: prevUuid } = {}, complete }) => {
                if (prevUuid !== uuid || complete) {
                    this.unregisterBeforeRender(uuid);
                }
            };
        }));
        this.updateSubscribers = this.effect(tap(({ animations, priority }) => {
            if (!animations.size)
                return;
            const subscribers = Array.from(animations.values());
            subscribers.sort((a, b) => (a.priority || 0) - (b.priority || 0));
            this.set((state) => ({
                internal: Object.assign(Object.assign({}, state.internal), { subscribers }),
            }));
        }));
        this.set({
            ready: false,
            clock: new THREE.Clock(),
            frameloop: 'always',
            legacy: false,
            linear: false,
            flat: false,
            orthographic: false,
            shadows: false,
            controls: null,
            pointer: this.pointer,
            mouse: this.pointer,
            sceneRef: new Ref(),
            cameraRef: new Ref(),
            events: {
                priority: 1,
                enabled: true,
                connected: undefined,
                handlers: {},
                compute: (event, state) => {
                    // https://github.com/pmndrs/react-three-fiber/pull/782
                    // Events trigger outside of canvas when moved, use offsetX/Y by default and allow overrides
                    state.pointer.set((event.offsetX / state.size.width) * 2 - 1, -(event.offsetY / state.size.height) * 2 + 1);
                    state.raycaster.setFromCamera(state.pointer, state.camera);
                },
            },
            cameraOptions: {},
            glOptions: {},
            raycasterOptions: {},
            sceneOptions: {},
            // eslint-disable-next-line @typescript-eslint/no-empty-function
            pointerMissed: () => { },
            internal: {
                active: false,
                lastEvent: null,
                priority: 0,
                frames: 0,
                interaction: [],
                hovered: new Map(),
                subscribers: [],
                initialClick: [0, 0],
                initialHits: [],
                capturedMap: new Map(),
                animations: new Map(),
            },
            dpr: [1, 2],
            size: {
                width: clientWidth,
                height: clientHeight,
            },
            viewport: {
                initialDpr: devicePixelRatio || 1,
                dpr: devicePixelRatio || 1,
                width: clientWidth,
                height: clientHeight,
                aspect: clientWidth / clientHeight,
                distance: 0,
                factor: 0,
                getCurrentViewport: (camera = this.get((s) => s.camera), target = this.defaultTarget, size = this.get((s) => s.size)) => {
                    const { width, height } = size;
                    const aspect = width / height;
                    if (target instanceof THREE.Vector3)
                        this.tempTarget.copy(target);
                    else
                        this.tempTarget.set(...target);
                    const distance = camera
                        .getWorldPosition(this.position)
                        .distanceTo(this.tempTarget);
                    if (is.orthographic(camera)) {
                        return {
                            width: width / camera.zoom,
                            height: height / camera.zoom,
                            factor: 1,
                            distance,
                            aspect,
                        };
                    }
                    const fov = (camera.fov * Math.PI) / 180; // convert vertical fov to radians
                    const h = 2 * Math.tan(fov / 2) * distance; // visible height
                    const w = h * (width / height);
                    return {
                        width: w,
                        height: h,
                        factor: width / w,
                        distance,
                        aspect,
                    };
                },
            },
            performance: Object.assign(Object.assign({}, performanceOptions), { regress: () => {
                    this.zone.runOutsideAngular(() => {
                        const performance = this.get((s) => s.performance);
                        // Clear timeout
                        if (this.performanceTimeoutId)
                            clearTimeout(this.performanceTimeoutId);
                        // Set lower bound performance
                        if (performance.current !== performance.min) {
                            this.set((state) => ({
                                performance: Object.assign(Object.assign({}, state.performance), { current: performance.min }),
                            }));
                        }
                        // Go back to upper bound performance after a while unless something regresses meanwhile
                        this.performanceTimeoutId = setTimeout(() => this.set((state) => ({
                            performance: Object.assign(Object.assign({}, state.performance), { current: this.get((s) => s.performance.max) }),
                        })), performance.debounce);
                    });
                } }),
            previousState: parentStore === null || parentStore === void 0 ? void 0 : parentStore.get(),
        });
    }
    init(canvasElement, rootStateMap, invalidate, advance) {
        this.initEvents(canvasElement);
        this.resize(this.resizeResult$);
        this.updateDimensions(this.dimensions$);
        this.initRenderer({ canvasElement, rootStateMap, advance });
        this.updateSubscribers(this.select(this.select((s) => s.internal.animations), this.select((s) => s.internal.priority), (animations, priority) => ({ animations, priority })));
        this.set((state) => ({
            invalidate: () => invalidate(() => this.get()),
            advance: (timestamp, runGlobalCallbacks) => advance(timestamp, runGlobalCallbacks, () => this.get()),
            internal: Object.assign(Object.assign({}, state.internal), { active: true }),
        }));
        this.set(this.allConstructed$);
    }
    registerBeforeRender(record) {
        const uuid = is.object3d(record.obj)
            ? record.obj.uuid
            : is.ref(record.obj)
                ? record.obj.value.uuid
                : makeId();
        this.registerBeforeRenderEffect(Object.assign(Object.assign({}, record), { uuid }));
        return () => {
            this.unregisterBeforeRender(uuid);
        };
    }
    unregisterBeforeRender(uuid) {
        if (!uuid)
            return;
        const currentAnimations = this.get((s) => s.internal.animations);
        const record = currentAnimations.get(uuid);
        const deleted = currentAnimations.delete(uuid);
        if (deleted && record) {
            this.set((state) => ({
                internal: Object.assign(Object.assign({}, state.internal), { animations: currentAnimations, priority: state.internal.priority -
                        ((record.priority || 0) > 0 ? 1 : 0) }),
            }));
        }
    }
    addInteraction(interaction) {
        this.set((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: [...state.internal.interaction, interaction] }) })));
    }
    removeInteraction(uuid) {
        this.set((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: state.internal.interaction.filter((interaction) => interaction.uuid !== uuid) }) })));
    }
    initEvents(canvasElement) {
        const { handlePointer } = createEvents(() => this.get());
        this.set((state) => ({
            events: Object.assign(Object.assign({}, state.events), { handlers: Object.keys(DOM_EVENTS).reduce((handlers, supportedEventName) => {
                    handlers[supportedEventName] =
                        handlePointer(supportedEventName);
                    return handlers;
                }, {}) }),
        }));
        this.connectElement(canvasElement);
    }
    connectElement(canvasElement) {
        this.set((state) => ({
            events: Object.assign(Object.assign({}, state.events), { connected: canvasElement }),
        }));
        const handlers = this.get((s) => s.events.handlers);
        Object.entries(handlers !== null && handlers !== void 0 ? handlers : {}).forEach(([eventName, handler]) => {
            const passive = DOM_EVENTS[eventName];
            canvasElement.addEventListener(eventName, handler, { passive });
        });
    }
    disconnectElement() {
        const { handlers, connected } = this.get((s) => s.events);
        if (connected) {
            Object.entries(handlers !== null && handlers !== void 0 ? handlers : {}).forEach(([eventName, handler]) => {
                if (connected instanceof HTMLElement) {
                    connected.removeEventListener(eventName, handler);
                }
            });
        }
    }
    ngOnDestroy() {
        this.disconnectElement();
        super.ngOnDestroy();
    }
}
NgtStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtStore, deps: [{ token: i0.ElementRef }, { token: NGT_PERFORMANCE_OPTIONS }, { token: WINDOW }, { token: NgtStore, optional: true, skipSelf: true }, { token: NgtResize }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Injectable });
NgtStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: i0.ElementRef }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_PERFORMANCE_OPTIONS]
                    }] }, { type: Window, decorators: [{
                        type: Inject,
                        args: [WINDOW]
                    }] }, { type: NgtStore, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }] }, { type: i1.Observable, decorators: [{
                        type: Inject,
                        args: [NgtResize]
                    }] }, { type: i0.NgZone }];
    } });
function createRenderer(glOptions, canvasElement) {
    const customRenderer = (typeof glOptions === 'function' ? glOptions(canvasElement) : glOptions);
    if ((customRenderer === null || customRenderer === void 0 ? void 0 : customRenderer.render) != null) {
        return customRenderer;
    }
    return new THREE.WebGLRenderer(Object.assign({ powerPreference: 'high-performance', canvas: canvasElement, antialias: true, alpha: true }, glOptions));
}

class NgtInstance extends NgtComponentStore {
    constructor(zone, store, parentRef, parentHostRef) {
        super();
        this.zone = zone;
        this.store = store;
        this.parentRef = parentRef;
        this.parentHostRef = parentHostRef;
        this.ready = new EventEmitter();
        this.hasEmittedAlready = false;
        this.instance$ = this.select((s) => s.instance).pipe(switchMap((instance) => instance.pipe(filter((instance) => instance != null))));
        this.instanceArgs$ = this.select((s) => s.instanceArgs);
        this.isPrimitive = false;
        this.isWrapper = false;
        this.isRaw = false;
        this.instanceReady = this.effect(tapEffect(() => {
            // assigning
            const setOptionsSub = this.setOptions(this.select(this.options$, this.setOptionsTrigger$, (options) => options));
            // attaching
            if (!this.get((s) => s.noAttach)) {
                this.attachToParent();
            }
            return () => {
                setOptionsSub.unsubscribe();
            };
        }));
        this.setOptions = this.effect(tap((options) => {
            // no options; return early
            if (Object.keys(options).length === 0)
                return;
            if (this.instanceValue) {
                // TODO: Material is handling this on their own. To be changed when [parameters] is removed
                if (is.material(this.instanceValue))
                    return;
                const state = this.get();
                const customOptions = {};
                const { rotation, quaternion } = options, restOptions = __rest(options, ["rotation", "quaternion"]);
                if (rotation) {
                    customOptions['rotation'] = state['rotation'];
                }
                else if (quaternion) {
                    customOptions['quaternion'] = state['quaternion'];
                }
                for (const option of Object.keys(restOptions)) {
                    if (state[option] != null) {
                        customOptions[option] = state[option];
                    }
                }
                applyProps(this.instanceValue, customOptions);
                this.postSetOptions(this.instanceValue);
                this.checkUpdate(this.instanceValue);
            }
        }));
        this.attachToParent = this.effect(pipe(withLatestFrom(this.select((s) => s.attach)), tap(([, attach]) => {
            var _a;
            let parentInstanceRef = this.__ngt__.parent;
            // if no parentInstance, try re-run the factory due to late init
            if (!parentInstanceRef || !parentInstanceRef.value) {
                // return early if failed to retrieve
                if (!((_a = this.parent) === null || _a === void 0 ? void 0 : _a.value))
                    return;
                // reassign on instance internal state
                this.__ngt__.parent = this.parent;
                parentInstanceRef = this.parent;
            }
            if (typeof attach === 'function') {
                const attachCleanUp = attach(parentInstanceRef, this.instance);
                if (attachCleanUp) {
                    this.__ngt__.previousAttach = attachCleanUp;
                }
            }
            else {
                const propertyToAttach = [...attach];
                // if propertyToAttach is empty
                if (propertyToAttach.length === 0)
                    return;
                // array material handling
                if (propertyToAttach[0] === 'material' &&
                    propertyToAttach[1] &&
                    is.num(Number(propertyToAttach[1])) &&
                    is.material(this.instanceValue)) {
                    if (!is.arr(parentInstanceRef.value.material)) {
                        parentInstanceRef.value.material = [];
                    }
                }
                // retrieve the current value on the parentInstance so we can reset it later
                this.__ngt__.previousAttachValue = propertyToAttach.reduce((value, property) => value[property], parentInstanceRef.value);
                // attach the instance value on the parent
                mutate(parentInstanceRef.value, propertyToAttach, this.instanceValue);
                // validate on the instance
                if (this.__ngt__) {
                    this.__ngt__.root().invalidate();
                }
                // also validate on the parentInstance
                if (parentInstanceRef.value.__ngt__) {
                    parentInstanceRef.value.__ngt__.root().invalidate();
                }
                this.__ngt__.previousAttach = propertyToAttach;
                this.set({
                    attach: propertyToAttach,
                });
            }
            this.checkUpdate(parentInstanceRef.value);
            this.checkUpdate(this.instanceValue);
        })));
        this.set({
            instance: new Ref(null),
            instanceArgs: [],
            attach: [],
            noAttach: false,
            skipParent: false,
        });
    }
    set ref(ref) {
        this.set({ instance: ref });
    }
    set skipParent(skipParent) {
        this.set({
            attachExplicit: true,
            skipParent: coerceBooleanProperty(skipParent),
        });
    }
    get skipParent() {
        return this.get((s) => s.skipParent);
    }
    set noAttach(noAttach) {
        this.set({
            noAttachExplicit: true,
            noAttach: coerceBooleanProperty(noAttach),
        });
    }
    get noAttach() {
        return this.get((s) => s.noAttach);
    }
    set attach(value) {
        if (value) {
            this.set({
                skipParentExplicit: true,
                attach: typeof value === 'function'
                    ? value
                    : is.arr(value)
                        ? value
                        : [value],
            });
        }
    }
    get attach() {
        return this.get((s) => s.attach);
    }
    get instance() {
        return this.get((s) => s.instance);
    }
    get instanceValue() {
        return this.isRaw
            ? this.instance.value.valueOf()
            : this.instance.value;
    }
    get __ngt__() {
        return this.instance.value.__ngt__;
    }
    get parent() {
        var _a, _b, _c;
        if (!this.skipParent)
            return (_a = this.parentRef) === null || _a === void 0 ? void 0 : _a.call(this);
        return ((_b = this.parentHostRef) === null || _b === void 0 ? void 0 : _b.call(this)) || ((_c = this.parentRef) === null || _c === void 0 ? void 0 : _c.call(this));
    }
    set instanceArgs(v) {
        this.set({
            instanceArgs: is.arr(v) ? v : [v],
        });
    }
    get instanceArgs() {
        return this.get((s) => s.instanceArgs);
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            this.preInit();
            this.onCanvasReady(this.store.ready$, () => {
                // make sure `instance()` is available before doing anything
                this.instanceReady(this.instance$);
            });
        });
    }
    /**
     * Sub-classes can use this function to run pre-init logic
     * @protected
     */
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    preInit() { }
    /**
     * Sub-classes can choose to run additional logic after init
     * @protected
     */
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    postInit() { }
    emitReady() {
        // only emit ready once to prevent reconstruct
        if (!this.hasEmittedAlready) {
            this.ready.emit(this.instanceValue);
            this.hasEmittedAlready = true;
        }
    }
    prepareInstance(instance, uuid) {
        var _a, _b, _c;
        if (uuid && 'uuid' in instance) {
            instance['uuid'] = uuid;
        }
        const prepInstance = prepare(instance, () => this.store.get(), this.parent, this.instance, this.isPrimitive);
        this.postPrepare(prepInstance);
        this.instance.set(prepInstance);
        this.emitReady();
        if (!is.object3d(prepInstance) && !this.noAttach && !this.skipParent) {
            const parentObjects = (_c = (_b = (_a = prepInstance.__ngt__.parent) === null || _a === void 0 ? void 0 : _a.value) === null || _b === void 0 ? void 0 : _b.__ngt__) === null || _c === void 0 ? void 0 : _c.objects;
            parentObjects &&
                parentObjects.set([...parentObjects.value, this.instance]);
        }
        return prepInstance;
    }
    /**
     * Sub-classes can use this to modify the constructor parameters
     * before calling this.prepareInstance
     */
    adjustCtorParams(instanceArgs) {
        return instanceArgs;
    }
    /**
     * Sub-classes, if adjust CtorParams, can also use ctorParams$ to ensure ctor is re-invoked
     */
    get ctorParams$() {
        return this.instanceArgs$.pipe(map(() => ({})));
    }
    destroy() {
        if (this.instanceValue) {
            if (is.object3d(this.instanceValue)) {
                const parentInstance = this.parent;
                if (parentInstance && is.object3d(parentInstance.value)) {
                    removeInteractivity(this.__ngt__.root.bind(this.__ngt__), this.instanceValue);
                }
                if (this.instanceValue.clear != null) {
                    this.instanceValue.clear();
                }
            }
            else {
                // non-scene objects
                const previousAttach = this.__ngt__.previousAttach;
                if (previousAttach != null) {
                    if (typeof previousAttach === 'function') {
                        previousAttach();
                        if (this.__ngt__.parent && this.__ngt__.parent.value) {
                            checkNeedsUpdate(this.__ngt__.parent.value);
                        }
                    }
                    else {
                        const previousAttachValue = this.__ngt__.previousAttachValue;
                        if (this.__ngt__.parent && this.__ngt__.parent.value) {
                            mutate(this.__ngt__.parent.value, previousAttach, previousAttachValue);
                            checkNeedsUpdate(this.__ngt__.parent.value);
                        }
                    }
                }
            }
            const dispose = this.instanceValue['dispose'];
            if (dispose && typeof dispose === 'function') {
                dispose.apply(this.instanceValue);
            }
        }
        this.set({ attach: [] });
        this.instance.complete();
    }
    /**
     * Can be used by sub-classes to run additional logic after prepare
     * @protected
     */
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    postPrepare(_) { }
    get optionFields() {
        return {};
    }
    get options$() {
        const optionEntries = Object.entries(this.optionFields);
        if (optionEntries.length === 0)
            return of({});
        return this.select(...optionEntries.map(([inputKey, shouldStartWithUndefined]) => {
            const subInput$ = this.select((s) => s[inputKey]);
            if (shouldStartWithUndefined)
                return subInput$.pipe(startWithUndefined());
            return subInput$;
        }), (...args) => args.reduce((record, arg, index) => {
            record[optionEntries[index][0]] = arg;
            return record;
        }, {})).pipe(startWith({}), pairwise(), map(([prev, curr]) => {
            return Object.entries(curr).reduce((options, [currKey, currValue]) => {
                if (!is.equ(prev[currKey], currValue)) {
                    options[currKey] = currValue;
                }
                return options;
            }, {});
        }));
    }
    ngOnDestroy() {
        this.zone.runOutsideAngular(() => {
            this.destroy();
        });
        super.ngOnDestroy();
    }
    /**
     * Can be used by sub-classes to run additional logic after options are set
     * @protected
     */
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    postSetOptions(_) { }
    /**
     * Sub-classes can customize this to run setOptions on custom trigger
     * @protected
     */
    get setOptionsTrigger$() {
        return of({});
    }
    checkUpdate(value) {
        if (is.object3d(value)) {
            value.updateMatrix();
        }
        else if (is.camera(value)) {
            if (is.perspective(value) || is.orthographic(value)) {
                value.updateProjectionMatrix();
            }
            value.updateMatrixWorld();
        }
        checkNeedsUpdate(value);
    }
}
NgtInstance.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtInstance, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_INSTANCE_REF, optional: true, skipSelf: true }, { token: NGT_INSTANCE_HOST_REF, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtInstance.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtInstance, inputs: { ref: "ref", skipParent: "skipParent", noAttach: "noAttach", attach: "attach" }, outputs: { ready: "ready" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtInstance, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_HOST_REF]
                    }] }];
    }, propDecorators: { ref: [{
                type: Input
            }], skipParent: [{
                type: Input
            }], ready: [{
                type: Output
            }], noAttach: [{
                type: Input
            }], attach: [{
                type: Input
            }] } });

const supportedEvents = [
    'click',
    'contextmenu',
    'dblclick',
    'pointerup',
    'pointerdown',
    'pointerover',
    'pointerout',
    'pointerenter',
    'pointerleave',
    'pointermove',
    'pointermissed',
    'pointercancel',
    'wheel',
];
class NgtObjectInputs extends NgtInstance {
    constructor(zone, store, parentObjectRef, parentObjectHostRef) {
        super(zone, store, parentObjectRef, parentObjectHostRef);
        this.parentObjectRef = parentObjectRef;
        this.parentObjectHostRef = parentObjectHostRef;
        // events
        this.click = new EventEmitter();
        this.contextmenu = new EventEmitter();
        this.dblclick = new EventEmitter();
        this.pointerup = new EventEmitter();
        this.pointerdown = new EventEmitter();
        this.pointerover = new EventEmitter();
        this.pointerout = new EventEmitter();
        this.pointerenter = new EventEmitter();
        this.pointerleave = new EventEmitter();
        this.pointermove = new EventEmitter();
        this.pointermissed = new EventEmitter();
        this.pointercancel = new EventEmitter();
        this.wheel = new EventEmitter();
        this.shouldPassThroughRef = true;
        this.set({
            name: '',
            position: new THREE.Vector3(),
            rotation: new THREE.Euler(),
            quaternion: new THREE.Quaternion(),
            scale: new THREE.Vector3(1, 1, 1),
            color: new THREE.Color(),
            castShadow: false,
            receiveShadow: false,
            priority: 0,
            visible: true,
            matrixAutoUpdate: true,
            appendMode: 'immediate',
            userData: {},
        });
    }
    set name(name) {
        this.set({
            nameExplicit: true,
            name,
        });
    }
    get name() {
        return this.get((s) => s.name);
    }
    set position(position) {
        this.set({
            positionExplicit: true,
            position: makeVector3(position),
        });
    }
    get position() {
        return this.get((s) => s.position);
    }
    get rotation() {
        return this.get((s) => s.rotation);
    }
    set rotation(rotation) {
        this.set({
            rotationExplicit: true,
            rotation: make(THREE.Euler, rotation),
        });
    }
    get quaternion() {
        return this.get((s) => s.quaternion);
    }
    set quaternion(quaternion) {
        this.set({
            quaternionExplicit: true,
            quaternion: make(THREE.Quaternion, quaternion),
        });
    }
    get scale() {
        return this.get((s) => s.scale);
    }
    set scale(scale) {
        this.set({
            scaleExplicit: true,
            scale: makeVector3(scale),
        });
    }
    get color() {
        return this.get((s) => s.color);
    }
    set color(color) {
        this.set({
            colorExplicit: true,
            color: makeColor(color),
        });
    }
    get castShadow() {
        return this.get((s) => s.castShadow);
    }
    set castShadow(value) {
        this.set({
            castShadowExplicit: true,
            castShadow: coerceBooleanProperty(value),
        });
    }
    get receiveShadow() {
        return this.get((s) => s.receiveShadow);
    }
    set receiveShadow(value) {
        this.set({
            receiveShadowExplicit: true,
            receiveShadow: coerceBooleanProperty(value),
        });
    }
    get priority() {
        return this.get((s) => s.priority);
    }
    set priority(priority) {
        this.set({
            priorityExplicit: true,
            priority: coerceNumberProperty(priority),
        });
    }
    get visible() {
        return this.get((s) => s.visible);
    }
    set visible(visible) {
        this.set({
            visibleExplicit: true,
            visible: coerceBooleanProperty(visible),
        });
    }
    get matrixAutoUpdate() {
        return this.get((s) => s.matrixAutoUpdate);
    }
    set matrixAutoUpdate(matrixAutoUpdate) {
        this.set({
            matrixAutoUpdateExplicit: true,
            matrixAutoUpdate: coerceBooleanProperty(matrixAutoUpdate),
        });
    }
    get userData() {
        return this.get((s) => s.userData);
    }
    set userData(userData) {
        this.set({
            userDataExplicit: true,
            userData,
        });
    }
    get dispose() {
        return this.get((s) => s.dispose);
    }
    set dispose(dispose) {
        this.set({
            disposeExplicit: true,
            dispose,
        });
    }
    get raycast() {
        return this.get((s) => s.raycast);
    }
    set raycast(raycast) {
        this.set({
            raycastExplicit: true,
            raycast,
        });
    }
    get appendMode() {
        return this.get((s) => s.appendMode);
    }
    set appendMode(appendMode) {
        this.set({
            appendModeExplicit: true,
            appendMode,
        });
    }
    get appendTo() {
        return this.get((s) => s.appendTo);
    }
    set appendTo(appendTo) {
        this.set({
            appendToExplicit: true,
            appendTo: appendTo
                ? is.ref(appendTo)
                    ? appendTo
                    : new Ref(appendTo)
                : undefined,
        });
    }
    get optionFields() {
        return {
            name: false,
            position: false,
            rotation: false,
            quaternion: false,
            scale: false,
            color: false,
            castShadow: false,
            receiveShadow: false,
            visible: false,
            matrixAutoUpdate: false,
            userData: false,
            dispose: true,
            raycast: true,
        };
    }
}
NgtObjectInputs.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectInputs, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_OBJECT_REF, optional: true, skipSelf: true }, { token: NGT_OBJECT_HOST_REF, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtObjectInputs.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtObjectInputs, inputs: { name: "name", position: "position", rotation: "rotation", quaternion: "quaternion", scale: "scale", color: "color", castShadow: "castShadow", receiveShadow: "receiveShadow", priority: "priority", visible: "visible", matrixAutoUpdate: "matrixAutoUpdate", userData: "userData", dispose: "dispose", raycast: "raycast", appendMode: "appendMode", appendTo: "appendTo" }, outputs: { click: "click", contextmenu: "contextmenu", dblclick: "dblclick", pointerup: "pointerup", pointerdown: "pointerdown", pointerover: "pointerover", pointerout: "pointerout", pointerenter: "pointerenter", pointerleave: "pointerleave", pointermove: "pointermove", pointermissed: "pointermissed", pointercancel: "pointercancel", wheel: "wheel" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectInputs, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_HOST_REF]
                    }] }];
    }, propDecorators: { name: [{
                type: Input
            }], position: [{
                type: Input
            }], rotation: [{
                type: Input
            }], quaternion: [{
                type: Input
            }], scale: [{
                type: Input
            }], color: [{
                type: Input
            }], castShadow: [{
                type: Input
            }], receiveShadow: [{
                type: Input
            }], priority: [{
                type: Input
            }], visible: [{
                type: Input
            }], matrixAutoUpdate: [{
                type: Input
            }], userData: [{
                type: Input
            }], dispose: [{
                type: Input
            }], raycast: [{
                type: Input
            }], appendMode: [{
                type: Input
            }], appendTo: [{
                type: Input
            }], click: [{
                type: Output
            }], contextmenu: [{
                type: Output
            }], dblclick: [{
                type: Output
            }], pointerup: [{
                type: Output
            }], pointerdown: [{
                type: Output
            }], pointerover: [{
                type: Output
            }], pointerout: [{
                type: Output
            }], pointerenter: [{
                type: Output
            }], pointerleave: [{
                type: Output
            }], pointermove: [{
                type: Output
            }], pointermissed: [{
                type: Output
            }], pointercancel: [{
                type: Output
            }], wheel: [{
                type: Output
            }] } });
class NgtObject extends NgtObjectInputs {
    constructor() {
        super(...arguments);
        this.appended = new EventEmitter();
        this.beforeRender = new EventEmitter();
        /**
         * @deprecated Use {@link beforeRender} instead
         */
        this.animateReady = this.beforeRender;
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                if (this.preObjectInit) {
                    this.preObjectInit(() => this.init());
                }
                else {
                    this.init();
                }
            });
        });
    }
    /**
     * Sub-classes can use this function to add additional logic BEFORE
     * initializing the object3d. The actual `initFn` is passed in so
     * the implementor needs to call this initFn() manually.
     *
     * This function is also called outside of Angular Zone
     * @protected
     */
    get preObjectInit() {
        return undefined;
    }
    init() {
        var _a, _b;
        if (this.instance.value && this.__ngt__) {
            this.switch();
        }
        else {
            this.prepareInstance(this.objectInitFn(), (_b = (_a = this.instance) === null || _a === void 0 ? void 0 : _a.value) === null || _b === void 0 ? void 0 : _b.uuid);
        }
        if (this.instance.value) {
            if (this.postInit) {
                this.postInit();
            }
            const observedEvents = supportedEvents.reduce((result, event) => {
                const controllerEvent = this[event].observed
                    ? this[event]
                    : null;
                if (controllerEvent) {
                    result.handlers[event] = this.eventNameToHandler(controllerEvent);
                    result.eventCount += 1;
                }
                return result;
            }, { handlers: {}, eventCount: 0 });
            // patch __ngt__ with events
            applyProps(this.__ngt__, {
                handlers: observedEvents.handlers,
                eventCount: observedEvents.eventCount,
            });
            // add as an interaction if there are events observed
            if (observedEvents.eventCount > 0) {
                this.store.addInteraction(this.instance.value);
            }
            // append to parent
            if (is.object3d(this.instance.value)) {
                this.appendToParent();
            }
            // setup beforeRender
            if (this.beforeRender.observed) {
                this.store.registerBeforeRender({
                    obj: this.instance,
                    callback: (state) => {
                        this.beforeRender.emit({
                            state,
                            object: this.instance.value,
                        });
                    },
                    priority: this.get((s) => s.priority),
                });
            }
        }
    }
    destroy() {
        if (this.instance.value) {
            // remove beforeRender callback
            this.store.unregisterBeforeRender(this.instance.value.uuid);
            // remove interaction
            if (this.__ngt__.eventCount > 0) {
                this.store.removeInteraction(this.instance.value.uuid);
            }
            this.remove();
        }
        super.destroy();
    }
    appendToParent() {
        // appendToParent is late a frame due to appendTo
        // only emit the object is ready after it's been added to the scene
        requestAnimationFrame(() => {
            const appendToRef = this.get((s) => s.appendTo);
            if (appendToRef && appendToRef.value) {
                appendToRef.value.add(this.instance.value);
                this.appended.emit(this.instance.value);
                return;
            }
            const appendMode = this.get((s) => s.appendMode);
            if (appendMode === 'none')
                return;
            if (appendMode === 'root') {
                this.addToScene();
                this.appended.emit(this.instance.value);
                return;
            }
            if (appendMode === 'immediate') {
                this.addToParent();
                this.appended.emit(this.instance.value);
            }
        });
    }
    addToScene() {
        const scene = this.store.get((s) => s.scene);
        if (scene) {
            scene.add(this.instance.value);
        }
    }
    addToParent() {
        var _a;
        let parent = this.parent;
        if (this.shouldUseParent(parent)) {
            parent.value.add(this.instance.value);
            return;
        }
        parent = (_a = this.parentHostRef) === null || _a === void 0 ? void 0 : _a.call(this);
        if (this.shouldUseParent(parent)) {
            parent.value.add(this.instance.value);
            return;
        }
        this.addToScene();
    }
    shouldUseParent(parent) {
        return (parent &&
            parent.value &&
            parent.value.uuid !== this.instance.value.uuid);
    }
    remove() {
        const { appendMode, appendTo } = this.get();
        if (appendTo && appendTo.value) {
            appendTo.value.remove(this.instance.value);
        }
        else if (this.shouldUseParent(this.parent) &&
            appendMode === 'immediate') {
            this.parent.value.remove(this.instance.value);
        }
        else {
            const scene = this.store.get((s) => s.scene);
            if (scene) {
                scene.remove(this.instance.value);
            }
        }
    }
    switch() {
        const newObject3d = this.objectInitFn();
        if (this.instance.value.children) {
            this.instance.value.traverse((object) => {
                if (object !== this.instance.value &&
                    object.parent === this.instance.value) {
                    object.parent = newObject3d;
                }
            });
            this.instance.value.children = [];
        }
        if (this.__ngt__.eventCount > 0) {
            this.store.removeInteraction(this.instance.value.uuid);
        }
        this.remove();
        this.prepareInstance(newObject3d);
    }
    eventNameToHandler(controllerEvent) {
        return (event) => {
            // go back into Angular Zone so that state updates on these events trigger CD
            this.zone.run(() => {
                controllerEvent.emit(event);
            });
        };
    }
}
NgtObject.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObject, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtObject.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtObject, outputs: { appended: "appended", beforeRender: "beforeRender", animateReady: "animateReady" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObject, decorators: [{
            type: Directive
        }], propDecorators: { appended: [{
                type: Output
            }], beforeRender: [{
                type: Output
            }], animateReady: [{
                type: Output
            }] } });

class NgtMaterialGeometry extends NgtObject {
    set material(material) {
        this.set({ material });
    }
    set geometry(geometry) {
        this.set({ geometry });
    }
    set morphTargetInfluences(morphTargetInfluences) {
        this.set({ morphTargetInfluences });
    }
    set morphTargetDictionary(morphTargetDictionary) {
        this.set({ morphTargetDictionary });
    }
    objectInitFn() {
        const state = this.get();
        // this is the additional arguments to pass into the object constructor
        // eg: InstancedMesh has "count" -> objectArgs = [count]
        const objectArgs = this.argsKeys.reduce((args, argKey) => {
            args.push(state[argKey]);
            return args;
        }, []);
        const object = new this.objectType(is.ref(state.geometry) ? state.geometry.value : state.geometry, is.arr(state.material)
            ? state.material.map((m) => (is.ref(m) ? m.value : m))
            : is.ref(state.material)
                ? state.material.value
                : state.material, ...objectArgs);
        if (state.morphTargetDictionary && 'morphTargetDictionary' in object) {
            object['morphTargetDictionary'] =
                state.morphTargetDictionary;
        }
        if (state.morphTargetInfluences && 'morphTargetInfluences' in object) {
            object['morphTargetInfluences'] =
                state.morphTargetInfluences;
        }
        return object;
    }
    get argsKeys() {
        return [];
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { material: true, geometry: true, morphTargetInfluences: true, morphTargetDictionary: true });
    }
}
NgtMaterialGeometry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMaterialGeometry, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtMaterialGeometry.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtMaterialGeometry, inputs: { material: "material", geometry: "geometry", morphTargetInfluences: "morphTargetInfluences", morphTargetDictionary: "morphTargetDictionary" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMaterialGeometry, decorators: [{
            type: Directive
        }], propDecorators: { material: [{
                type: Input
            }], geometry: [{
                type: Input
            }], morphTargetInfluences: [{
                type: Input
            }], morphTargetDictionary: [{
                type: Input
            }] } });

function provideInstanceRef(subType, factory) {
    return [
        { provide: NgtInstance, useExisting: subType },
        {
            provide: NGT_INSTANCE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subType],
        },
    ];
}
function provideInstanceHostRef(subType, factory, hostFactory) {
    return [
        {
            provide: NGT_INSTANCE_REF,
            useFactory: (instance) => {
                return () => factory(instance);
            },
            deps: [subType],
        },
        {
            provide: NGT_INSTANCE_HOST_REF,
            useFactory: (instance) => {
                return hostFactory
                    ? hostFactory(instance)
                    : () => instance.parent;
            },
            deps: [subType],
        },
    ];
}

function provideObjectRef(subType, factory) {
    return [
        provideInstanceRef(subType, factory),
        { provide: NgtObject, useExisting: subType },
        {
            provide: NGT_OBJECT_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subType],
        },
    ];
}
function provideObjectHostRef(subType, factory, hostFactory) {
    return [
        { provide: NgtObject, useExisting: subType },
        {
            provide: NGT_INSTANCE_REF,
            useFactory: (instance) => {
                return () => factory ? factory(instance) : instance['instance'];
            },
            deps: [subType],
        },
        {
            provide: NGT_INSTANCE_HOST_REF,
            useFactory: (instance) => {
                return hostFactory ? hostFactory(instance) : instance.parentRef;
            },
            deps: [subType],
        },
        {
            provide: NGT_OBJECT_REF,
            useFactory: (instance) => {
                return () => factory ? factory(instance) : instance['instance'];
            },
            deps: [subType],
        },
        {
            provide: NGT_OBJECT_HOST_REF,
            useFactory: (instance) => {
                return hostFactory ? hostFactory(instance) : instance.parentRef;
            },
            deps: [subType],
        },
    ];
}
function provideCanvasInstanceRef(canvasType) {
    return [
        {
            provide: NGT_INSTANCE_REF,
            useFactory: (canvas) => {
                return () => canvas.sceneRef;
            },
            deps: [canvasType],
        },
        {
            provide: NGT_OBJECT_REF,
            useFactory: (canvas) => {
                return () => canvas.sceneRef;
            },
            deps: [canvasType],
        },
        {
            provide: NGT_SCENE_REF,
            useFactory: (canvas) => {
                return () => canvas.sceneRef;
            },
            deps: [canvasType],
        },
        {
            provide: NGT_CAMERA_REF,
            useFactory: (canvas) => {
                return () => canvas.cameraRef;
            },
            deps: [canvasType],
        },
    ];
}

const globalCallbacks = [];
const globalAfterCallbacks = [];
const globalTailCallbacks = [];
function createCallback(callback, callbacks) {
    const index = callbacks.length;
    callbacks.push(callback);
    return () => void callbacks.splice(index, 1);
}
function runCallbacks(effects, timestamp) {
    for (let i = 0, length = effects.length; i < length; i++) {
        effects[i](timestamp);
    }
}
/**
 * Adds a global render callback which is called before each frame.
 */
function addCallback(callback) {
    return createCallback(callback, globalCallbacks);
}
/**
 * Adds a global render callback which is called after each frame has been rendered.
 */
function addAfterCallback(callback) {
    return createCallback(callback, globalAfterCallbacks);
}
/**
 * Adds a global callback which is called when rendering stops.
 */
function addTailCallback(callback) {
    return createCallback(callback, globalTailCallbacks);
}
function render(timestamp, state, frame) {
    const rootState = state();
    let delta = rootState.clock.getDelta();
    // in frameloop='never' mode, clock times are updated using provided timestamp
    if (rootState.frameloop === 'never' && typeof timestamp === 'number') {
        delta = timestamp - rootState.clock.elapsedTime;
        rootState.clock.oldTime = rootState.clock.elapsedTime;
        rootState.clock.elapsedTime = timestamp;
    }
    // call animation subscribers
    for (const subscriber of rootState.internal.subscribers) {
        const object = is.ref(subscriber.obj)
            ? subscriber.obj.value
            : subscriber.obj;
        subscriber.callback(Object.assign(Object.assign({}, rootState), { delta, frame }), object);
    }
    // Render content
    if (!rootState.internal.priority && rootState.gl.render) {
        rootState.gl.render(rootState.scene, rootState.camera);
    }
    // Decrease frame count
    rootState.internal.frames = Math.max(0, rootState.internal.frames - 1);
    return rootState.frameloop === 'always' ? 1 : rootState.internal.frames;
}
function createLoop(rootStateMap) {
    let running = false;
    let repeat;
    let frame;
    let state;
    function loop(timestamp) {
        frame = requestAnimationFrame(loop);
        running = true;
        repeat = 0;
        // Run callbacks
        if (globalCallbacks.length) {
            runCallbacks(globalCallbacks, timestamp);
        }
        // Render all roots
        rootStateMap.forEach((rootState) => {
            var _a;
            state = rootState();
            // If the frameloop is invalidated, do not run another frame
            if (state.internal.active &&
                (state.frameloop === 'always' || state.internal.frames > 0) &&
                !((_a = state.gl.xr) === null || _a === void 0 ? void 0 : _a.isPresenting)) {
                repeat += render(timestamp, rootState);
            }
        });
        // Run after-callbacks
        if (globalAfterCallbacks.length) {
            runCallbacks(globalAfterCallbacks, timestamp);
        }
        // Stop the loop if nothing invalidates it
        if (repeat === 0) {
            // Tail call effects, they are called when rendering stops
            if (globalTailCallbacks.length) {
                runCallbacks(globalTailCallbacks, timestamp);
            }
            // Flag end of operation
            running = false;
            return cancelAnimationFrame(frame);
        }
    }
    function invalidate(state) {
        var _a;
        const stateToInvalidate = state === null || state === void 0 ? void 0 : state();
        if (!stateToInvalidate)
            return rootStateMap.forEach((rootState) => invalidate(rootState));
        if (((_a = stateToInvalidate.gl.xr) === null || _a === void 0 ? void 0 : _a.isPresenting) ||
            !stateToInvalidate.internal.active ||
            stateToInvalidate.frameloop === 'never')
            return;
        // Increase frames, do not go higher than 60
        stateToInvalidate.internal.frames = Math.min(60, stateToInvalidate.internal.frames + 1);
        // If the render-loop isn't active, start it
        if (!running) {
            running = true;
            requestAnimationFrame(loop);
        }
    }
    function advance(timestamp, runGlobalCallbacks = true, state, frame) {
        if (runGlobalCallbacks)
            runCallbacks(globalCallbacks, timestamp);
        const stateToAdvance = state === null || state === void 0 ? void 0 : state();
        if (!stateToAdvance) {
            rootStateMap.forEach((rootState) => render(timestamp, rootState));
        }
        else {
            render(timestamp, () => stateToAdvance, frame);
        }
        if (runGlobalCallbacks)
            runCallbacks(globalAfterCallbacks, timestamp);
    }
    return { loop, advance, invalidate };
}

const rootStateMap = new Map();
const { invalidate, advance } = createLoop(rootStateMap);
class NgtCanvas extends NgtComponentStore {
    constructor(zone, store) {
        super();
        this.zone = zone;
        this.store = store;
        this.hostClass = true;
        this._initialLog = false;
        this._projectContent = false;
        this.created = new EventEmitter();
        this.pointermissed = new EventEmitter();
    }
    set linear(linear) {
        this.store.set({ linear: coerceBooleanProperty(linear) });
    }
    set legacy(legacy) {
        this.store.set({ legacy: coerceBooleanProperty(legacy) });
    }
    set flat(flat) {
        this.store.set({ flat: coerceBooleanProperty(flat) });
    }
    set frameloop(frameloop) {
        this.store.set({ frameloop });
    }
    set orthographic(orthographic) {
        this.store.set({
            orthographic: coerceBooleanProperty(orthographic),
        });
    }
    set dpr(dpr) {
        this.store.set({ dpr });
    }
    set raycaster(raycaster) {
        this.store.set({ raycasterOptions: raycaster });
    }
    set shadows(shadows) {
        this.store.set({
            shadows: typeof shadows === 'object'
                ? shadows
                : coerceBooleanProperty(shadows),
        });
    }
    set camera(cameraOptions) {
        this.store.set({ cameraOptions });
    }
    get cameraRef() {
        return this.store.get((s) => s.cameraRef);
    }
    set scene(sceneOptions) {
        this.store.set({ sceneOptions });
    }
    get sceneRef() {
        return this.store.get((s) => s.sceneRef);
    }
    set gl(glOptions) {
        this.store.set({ glOptions });
    }
    get initialLog() {
        return this._initialLog;
    }
    set initialLog(value) {
        this._initialLog = coerceBooleanProperty(value);
    }
    get projectContent() {
        return this._projectContent;
    }
    set projectContent(value) {
        this._projectContent = coerceBooleanProperty(value);
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            // if there is handler to pointermissed on the canvas
            // update pointermissed in events store so that
            // events util will handle it
            if (this.pointermissed.observed) {
                this.store.set({
                    pointerMissed: (event) => {
                        this.pointermissed.emit(event);
                    },
                });
            }
            rootStateMap.set(this.rendererCanvas.nativeElement, () => this.store.get());
            this.store.init(this.rendererCanvas.nativeElement, rootStateMap, invalidate, advance);
            this.onCanvasReady(this.store.ready$, () => {
                const canvasState = this.store.get();
                this.created.emit(canvasState);
                if (this.initialLog) {
                    console.group('[NgtCanvas] Initialized');
                    console.log(canvasState);
                    console.groupEnd();
                }
                this.store.startLoop(this.store.select());
            });
        });
    }
}
NgtCanvas.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvas, deps: [{ token: i0.NgZone }, { token: NgtStore }], target: i0.ɵɵFactoryTarget.Component });
NgtCanvas.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "13.3.4", type: NgtCanvas, selector: "ngt-canvas", inputs: { linear: "linear", legacy: "legacy", flat: "flat", frameloop: "frameloop", orthographic: "orthographic", dpr: "dpr", raycaster: "raycaster", shadows: "shadows", camera: "camera", scene: "scene", gl: "gl", initialLog: "initialLog", projectContent: "projectContent" }, outputs: { created: "created", pointermissed: "pointermissed" }, host: { properties: { "class.ngt-canvas": "this.hostClass" } }, providers: [NgtStore, NgtResize, provideCanvasInstanceRef(NgtCanvas)], viewQueries: [{ propertyName: "rendererCanvas", first: true, predicate: ["rendererCanvas"], descendants: true, static: true }], usesInheritance: true, ngImport: i0, template: `
        <canvas #rendererCanvas></canvas>
        <ng-container
            *ngIf="projectContent"
            [ngTemplateOutlet]="contentTemplate"
        ></ng-container>
        <ng-template #contentTemplate>
            <ng-content></ng-content>
        </ng-template>
    `, isInline: true, styles: [":host{display:block;position:relative;width:100%;height:100%;overflow:hidden}:host canvas{display:block}\n"], directives: [{ type: i2.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { type: i2.NgTemplateOutlet, selector: "[ngTemplateOutlet]", inputs: ["ngTemplateOutletContext", "ngTemplateOutlet"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvas, decorators: [{
            type: Component,
            args: [{
                    selector: 'ngt-canvas',
                    template: `
        <canvas #rendererCanvas></canvas>
        <ng-container
            *ngIf="projectContent"
            [ngTemplateOutlet]="contentTemplate"
        ></ng-container>
        <ng-template #contentTemplate>
            <ng-content></ng-content>
        </ng-template>
    `,
                    changeDetection: ChangeDetectionStrategy.OnPush,
                    styles: [
                        `
            :host {
                display: block;
                position: relative;
                width: 100%;
                height: 100%;
                overflow: hidden;
            }

            :host canvas {
                display: block;
            }
        `,
                    ],
                    providers: [NgtStore, NgtResize, provideCanvasInstanceRef(NgtCanvas)],
                }]
        }], ctorParameters: function () { return [{ type: i0.NgZone }, { type: NgtStore }]; }, propDecorators: { hostClass: [{
                type: HostBinding,
                args: ['class.ngt-canvas']
            }], linear: [{
                type: Input
            }], legacy: [{
                type: Input
            }], flat: [{
                type: Input
            }], frameloop: [{
                type: Input
            }], orthographic: [{
                type: Input
            }], dpr: [{
                type: Input
            }], raycaster: [{
                type: Input
            }], shadows: [{
                type: Input
            }], camera: [{
                type: Input
            }], scene: [{
                type: Input
            }], gl: [{
                type: Input
            }], initialLog: [{
                type: Input
            }], projectContent: [{
                type: Input
            }], created: [{
                type: Output
            }], pointermissed: [{
                type: Output
            }], rendererCanvas: [{
                type: ViewChild,
                args: ['rendererCanvas', { static: true }]
            }] } });
class NgtCanvasModule {
}
NgtCanvasModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvasModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCanvasModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvasModule, declarations: [NgtCanvas], imports: [CommonModule], exports: [NgtCanvas] });
NgtCanvasModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvasModule, imports: [[CommonModule]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCanvasModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtCanvas],
                    exports: [NgtCanvas],
                    imports: [CommonModule],
                }]
        }] });
/**
 * @deprecated Use {@link NgtCanvasModule} instead. Will be removed in next major version
 */
class NgtCoreModule {
}
NgtCoreModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCoreModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCoreModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCoreModule, imports: [NgtCanvasModule], exports: [NgtCanvasModule] });
NgtCoreModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCoreModule, imports: [[NgtCanvasModule], NgtCanvasModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCoreModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [NgtCanvasModule],
                    exports: [NgtCanvasModule],
                }]
        }] });

const privateKeys = [
    'events',
    'invalidate',
    'advance',
    'size',
    'viewport',
];
class NgtPortalContent {
    constructor(templateRef) {
        this.templateRef = templateRef;
    }
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
}
NgtPortalContent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalContent, deps: [{ token: i0.TemplateRef }], target: i0.ɵɵFactoryTarget.Directive });
NgtPortalContent.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtPortalContent, selector: "ng-template[ngt-portal-content]", ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalContent, decorators: [{
            type: Directive,
            args: [{
                    selector: 'ng-template[ngt-portal-content]',
                }]
        }], ctorParameters: function () { return [{ type: i0.TemplateRef }]; } });
class NgtPortal extends NgtInstance {
    constructor(zone, store, parentRef, parentHostRef, resizeResult$) {
        super(zone, store, parentRef, parentHostRef);
        this.resizeResult$ = resizeResult$;
        this.init = this.effect(tap(() => {
            const previousState = this.store.get();
            this.set((state) => {
                var _a, _b;
                return ({
                    raycaster: new THREE.Raycaster(),
                    pointer: new THREE.Vector2(),
                    size: (_a = state.size) !== null && _a !== void 0 ? _a : previousState.size,
                    events: (_b = state.events) !== null && _b !== void 0 ? _b : previousState.events,
                });
            });
        }));
        this.initPortalStore = this.effect(tap(() => {
            var _a;
            const previousState = this.store.get();
            const portalState = this.get();
            const clientDimensions = {
                clientWidth: portalState.size.width,
                clientHeight: portalState.size.height,
            };
            const _b = previousState.performance, { regress: _ } = _b, performanceOptions = __rest(_b, ["regress"]);
            const portalStore = new NgtStore({ nativeElement: clientDimensions }, performanceOptions, {
                devicePixelRatio: makeDpr(previousState.viewport.dpr),
            }, this.store, this.resizeResult$, this.zone);
            const sceneRef = this.instance;
            let scene = this.instance.value;
            if (!scene) {
                scene = prepare(new THREE.Scene(), () => portalStore.get(), (_a = this.parentRef) === null || _a === void 0 ? void 0 : _a.call(this));
                this.instance.set(scene);
            }
            portalStore.set(Object.assign(Object.assign({}, previousState), { scene,
                sceneRef, raycaster: portalState.raycaster, pointer: portalState.pointer, mouse: portalState.pointer, previousState: this.store.get(), events: Object.assign(Object.assign({}, previousState.events), portalState.events), size: Object.assign(Object.assign({}, previousState.size), portalState.size) }));
            this.set({ portalStore });
        }));
        this.subscribeToRoot = this.effect(tapEffect(() => {
            const unsub = this.store.select().subscribe((root) => {
                this.get((s) => s.portalStore).set((portal) => this.inject(root, portal));
            });
            return () => {
                unsub.unsubscribe();
            };
        }));
        this.setInitialPortalState = this.effect(tap(() => {
            this.get((s) => s.portalStore).set((portal) => this.inject(this.store.get(), portal));
        }));
    }
    set size(size) {
        this.set({ size });
    }
    set events(events) {
        this.set({ events });
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            // portal initialization does not need to wait for canvas ready
            this.init();
            this.onCanvasReady(this.store.ready$, () => {
                // init portal store
                this.initPortalStore();
                this.subscribeToRoot();
                this.setInitialPortalState();
            });
        });
    }
    inject(root, injected) {
        const { size, events, portalStore } = this.get();
        const _a = portalStore.get(), { scene, raycaster, pointer, previousState, events: portalStoreEvents, size: portalStoreSize, viewport: portalStoreViewport, mouse } = _a, rest = __rest(_a, ["scene", "raycaster", "pointer", "previousState", "events", "size", "viewport", "mouse"]);
        const intersect = Object.assign({}, root);
        Object.keys(root).forEach((key) => {
            if (privateKeys.includes(key) ||
                root[key] !== injected[key]) {
                delete intersect[key];
            }
        });
        let viewport = undefined;
        if (injected && size) {
            const camera = injected.camera;
            viewport = root.viewport.getCurrentViewport(camera, new THREE.Vector3(), size);
            if (camera !== root.camera) {
                updateCamera(root.cameraOptions, camera, size);
            }
        }
        return Object.assign(Object.assign(Object.assign({}, intersect), { scene,
            raycaster,
            pointer,
            mouse,
            previousState, events: Object.assign(Object.assign(Object.assign({}, root.events), ((injected === null || injected === void 0 ? void 0 : injected.events) || portalStoreEvents)), events), size: Object.assign(Object.assign(Object.assign({}, root.size), portalStoreSize), size), viewport: Object.assign(Object.assign({}, root.viewport), (viewport || portalStoreViewport)) }), rest);
    }
}
NgtPortal.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortal, deps: [{ token: i0.NgZone }, { token: NgtStore, skipSelf: true }, { token: NGT_INSTANCE_REF, optional: true, skipSelf: true }, { token: NGT_INSTANCE_HOST_REF, optional: true, skipSelf: true }, { token: NgtResize }], target: i0.ɵɵFactoryTarget.Component });
NgtPortal.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "13.3.4", type: NgtPortal, selector: "ngt-portal", inputs: { size: "size", events: "events" }, providers: [
        NgtResize,
        {
            provide: NgtStore,
            useFactory: (portal) => {
                return portal.get((s) => s.portalStore);
            },
            deps: [NgtPortal],
        },
        {
            provide: NGT_INSTANCE_REF,
            useFactory: (portal) => {
                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
            },
            deps: [NgtPortal],
        },
        {
            provide: NGT_OBJECT_REF,
            useFactory: (portal) => {
                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
            },
            deps: [NgtPortal],
        },
        {
            provide: NGT_SCENE_REF,
            useFactory: (portal) => {
                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
            },
            deps: [NgtPortal],
        },
        {
            provide: NGT_CAMERA_REF,
            useFactory: (portal) => {
                return () => portal.get((s) => s.portalStore).get((s) => s.cameraRef);
            },
            deps: [NgtPortal],
        },
    ], queries: [{ propertyName: "content", first: true, predicate: NgtPortalContent, descendants: true }], usesInheritance: true, ngImport: i0, template: `
        <ng-container
            *ngIf="content"
            [ngTemplateOutlet]="content.templateRef"
            [ngTemplateOutletContext]="{ portal: instance }"
        ></ng-container>
        <ng-content></ng-content>
    `, isInline: true, directives: [{ type: i2.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { type: i2.NgTemplateOutlet, selector: "[ngTemplateOutlet]", inputs: ["ngTemplateOutletContext", "ngTemplateOutlet"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortal, decorators: [{
            type: Component,
            args: [{
                    selector: 'ngt-portal',
                    template: `
        <ng-container
            *ngIf="content"
            [ngTemplateOutlet]="content.templateRef"
            [ngTemplateOutletContext]="{ portal: instance }"
        ></ng-container>
        <ng-content></ng-content>
    `,
                    changeDetection: ChangeDetectionStrategy.OnPush,
                    providers: [
                        NgtResize,
                        {
                            provide: NgtStore,
                            useFactory: (portal) => {
                                return portal.get((s) => s.portalStore);
                            },
                            deps: [NgtPortal],
                        },
                        {
                            provide: NGT_INSTANCE_REF,
                            useFactory: (portal) => {
                                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
                            },
                            deps: [NgtPortal],
                        },
                        {
                            provide: NGT_OBJECT_REF,
                            useFactory: (portal) => {
                                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
                            },
                            deps: [NgtPortal],
                        },
                        {
                            provide: NGT_SCENE_REF,
                            useFactory: (portal) => {
                                return () => portal.get((s) => s.portalStore).get((s) => s.sceneRef);
                            },
                            deps: [NgtPortal],
                        },
                        {
                            provide: NGT_CAMERA_REF,
                            useFactory: (portal) => {
                                return () => portal.get((s) => s.portalStore).get((s) => s.cameraRef);
                            },
                            deps: [NgtPortal],
                        },
                    ],
                }]
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore, decorators: [{
                        type: SkipSelf
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_HOST_REF]
                    }] }, { type: i1.Observable, decorators: [{
                        type: Inject,
                        args: [NgtResize]
                    }] }];
    }, propDecorators: { size: [{
                type: Input
            }], events: [{
                type: Input
            }], content: [{
                type: ContentChild,
                args: [NgtPortalContent]
            }] } });
class NgtPortalModule {
}
NgtPortalModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtPortalModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalModule, declarations: [NgtPortal, NgtPortalContent], imports: [CommonModule], exports: [NgtPortal, NgtPortalContent] });
NgtPortalModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalModule, imports: [[CommonModule]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPortalModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtPortal, NgtPortalContent],
                    exports: [NgtPortal, NgtPortalContent],
                    imports: [CommonModule],
                }]
        }] });

class NgtCommonAttribute extends NgtInstance {
    constructor() {
        super(...arguments);
        this.init = this.effect(tapEffect((instanceArgs) => {
            this.prepareInstance(new this.attributeType(...instanceArgs));
        }));
    }
    set args(v) {
        this.instanceArgs = v;
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                const initSub = this.init(this.instanceArgs$);
                this.postInit();
                return () => {
                    if (initSub && initSub.unsubscribe) {
                        initSub.unsubscribe();
                    }
                };
            }, true);
        });
    }
}
NgtCommonAttribute.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonAttribute, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonAttribute.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonAttribute, inputs: { args: "args" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonAttribute, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }] } });

function provideCommonAttributeRef(subAttributeType, factory) {
    return [
        provideInstanceRef(subAttributeType, factory),
        { provide: NgtCommonAttribute, useExisting: subAttributeType },
        {
            provide: NGT_COMMON_ATTRIBUTE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subAttributeType],
        },
    ];
}

class NgtCommonAudio extends NgtObject {
    set listener(listener) {
        this.set({ listener });
    }
    set buffer(buffer) {
        this.set({ buffer });
    }
    set detune(detune) {
        this.set({ detune: coerceNumberProperty(detune) });
    }
    set loop(loop) {
        this.set({ loop: coerceBooleanProperty(loop) });
    }
    set loopStart(loopStart) {
        this.set({ loopStart: coerceNumberProperty(loopStart) });
    }
    set loopEnd(loopEnd) {
        this.set({ loopEnd: coerceNumberProperty(loopEnd) });
    }
    set offset(offset) {
        this.set({ offset: coerceNumberProperty(offset) });
    }
    set playbackRate(playbackRate) {
        this.set({ playbackRate: coerceNumberProperty(playbackRate) });
    }
    set isPlaying(isPlaying) {
        this.set({ isPlaying: coerceBooleanProperty(isPlaying) });
    }
    set hasPlaybackControl(hasPlaybackControl) {
        this.set({
            hasPlaybackControl: coerceBooleanProperty(hasPlaybackControl),
        });
    }
    set sourceType(sourceType) {
        this.set({ sourceType });
    }
    set source(source) {
        this.set({ source });
    }
    set filters(filters) {
        this.set({ filters });
    }
    set duration(duration) {
        this.set({ duration: coerceNumberProperty(duration) });
    }
    set autoplay(autoplay) {
        this.set({ autoplay: coerceBooleanProperty(autoplay) });
    }
    objectInitFn() {
        const { listener } = this.get();
        if (!listener) {
            throw new Error(`[NgtCommonAudio] Cannot initialize ${this.audioType} without "THREE.AudioListener"`);
        }
        return new this.audioType(listener);
    }
    preInit() {
        this.set((state) => ({
            buffer: state.buffer || null,
            detune: state.detune || 0,
            loop: state.loop || false,
            loopStart: state.loopStart || 0,
            loopEnd: state.loopEnd || 0,
            offset: state.offset || 0,
            duration: state.duration || undefined,
            playbackRate: state.playbackRate || 1,
            isPlaying: state.isPlaying || false,
            hasPlaybackControl: state.hasPlaybackControl || true,
            sourceType: state.sourceType || 'empty',
            source: state.source || null,
            filters: state.filters || [],
        }));
    }
    get preObjectInit() {
        return (initFn) => {
            this.effect(tapEffect(() => {
                initFn();
            }))(this.select((s) => s.listener));
        };
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { autoplay: true, buffer: false, detune: false, loop: false, loopStart: false, loopEnd: false, offset: false, duration: true, playbackRate: false, isPlaying: false, hasPlaybackControl: false, sourceType: false, source: false, filters: false });
    }
}
NgtCommonAudio.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonAudio, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonAudio.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonAudio, inputs: { listener: "listener", buffer: "buffer", detune: "detune", loop: "loop", loopStart: "loopStart", loopEnd: "loopEnd", offset: "offset", playbackRate: "playbackRate", isPlaying: "isPlaying", hasPlaybackControl: "hasPlaybackControl", sourceType: "sourceType", source: "source", filters: "filters", duration: "duration", autoplay: "autoplay" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonAudio, decorators: [{
            type: Directive
        }], propDecorators: { listener: [{
                type: Input
            }], buffer: [{
                type: Input
            }], detune: [{
                type: Input
            }], loop: [{
                type: Input
            }], loopStart: [{
                type: Input
            }], loopEnd: [{
                type: Input
            }], offset: [{
                type: Input
            }], playbackRate: [{
                type: Input
            }], isPlaying: [{
                type: Input
            }], hasPlaybackControl: [{
                type: Input
            }], sourceType: [{
                type: Input
            }], source: [{
                type: Input
            }], filters: [{
                type: Input
            }], duration: [{
                type: Input
            }], autoplay: [{
                type: Input
            }] } });

function provideCommonAudioRef(subAudioType, factory) {
    return [
        provideObjectRef(subAudioType, factory),
        { provide: NgtCommonAudio, useExisting: subAudioType },
        {
            provide: NGT_COMMON_AUDIO_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subAudioType],
        },
    ];
}

class NgtCommonCamera extends NgtObject {
    set args(v) {
        this.instanceArgs = v;
    }
    objectInitFn() {
        const instanceArgs = this.get((s) => s.instanceArgs);
        return new this.cameraType(...instanceArgs);
    }
}
NgtCommonCamera.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonCamera, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonCamera.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonCamera, inputs: { args: "args" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonCamera, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }] } });

function provideCommonCameraRef(subCameraType, factory) {
    return [
        provideObjectRef(subCameraType, factory),
        { provide: NgtCommonCamera, useExisting: subCameraType },
        {
            provide: NGT_COMMON_CAMERA_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subCameraType],
        },
    ];
}

class NgtCommonCurve extends NgtInstance {
    constructor() {
        super(...arguments);
        this.init = this.effect(tapEffect((instanceArgs) => {
            this.prepareInstance(new this.curveType(...instanceArgs));
        }));
    }
    set args(v) {
        this.instanceArgs = v;
    }
    set arcLengthDivisions(arcLengthDivisions) {
        this.set({
            arcLengthDivisions: coerceNumberProperty(arcLengthDivisions),
        });
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                this.init(this.instanceArgs$);
                this.postInit();
            });
        });
    }
    postPrepare(curve) {
        const arcLengthDivisions = this.get((s) => s.arcLengthDivisions);
        if (arcLengthDivisions != undefined) {
            curve.arcLengthDivisions = arcLengthDivisions;
        }
    }
}
NgtCommonCurve.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonCurve, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonCurve.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonCurve, inputs: { args: "args", arcLengthDivisions: "arcLengthDivisions" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonCurve, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }], arcLengthDivisions: [{
                type: Input
            }] } });

function provideCommonCurveRef(subCurveType, factory) {
    return [
        provideInstanceRef(subCurveType, factory),
        { provide: NgtCommonCurve, useExisting: subCurveType },
        {
            provide: NGT_COMMON_CURVE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subCurveType],
        },
    ];
}

class NgtCommonGeometry extends NgtInstance {
    constructor(zone, store, parentRef, parentHostRef) {
        super(zone, store, parentRef, parentHostRef);
        this.init = this.effect(tapEffect(() => {
            const instanceArgs = this.get((s) => s.instanceArgs);
            const geometryArgs = this.adjustCtorParams(instanceArgs);
            const geometry = this.prepareInstance(new this.geometryType(...geometryArgs));
            return () => {
                geometry.dispose();
            };
        }));
    }
    set args(v) {
        this.instanceArgs = v;
    }
    preInit() {
        this.set((state) => ({
            attach: state.attach.length ? state.attach : ['geometry'],
        }));
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                this.init(this.ctorParams$);
                this.postInit();
            });
        });
    }
}
NgtCommonGeometry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonGeometry, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_OBJECT_REF, optional: true, skipSelf: true }, { token: NGT_OBJECT_HOST_REF, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonGeometry.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonGeometry, inputs: { args: "args" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonGeometry, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_HOST_REF]
                    }] }];
    }, propDecorators: { args: [{
                type: Input
            }] } });

function provideCommonGeometryRef(subGeometryType, factory) {
    return [
        provideInstanceRef(subGeometryType, factory),
        { provide: NgtCommonGeometry, useExisting: subGeometryType },
        {
            provide: NGT_COMMON_GEOMETRY_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subGeometryType],
        },
    ];
}

class NgtCommonHelper extends NgtObject {
    set args(v) {
        this.instanceArgs = v;
    }
    objectInitFn() {
        const instanceArgs = this.get((s) => s.instanceArgs);
        return new this.helperType(...instanceArgs);
    }
}
NgtCommonHelper.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonHelper, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonHelper.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonHelper, inputs: { args: "args" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonHelper, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }] } });

function provideCommonHelperRef(subHelperType, factory) {
    return [
        provideObjectRef(subHelperType, factory),
        { provide: NgtCommonHelper, useExisting: subHelperType },
        {
            provide: NGT_COMMON_HELPER_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subHelperType],
        },
    ];
}

const NGT_IS_WEBGL_AVAILABLE = new InjectionToken('isWebGLAvailable', {
    factory: () => {
        const document = inject(DOCUMENT);
        const window = inject(WINDOW);
        try {
            const canvas = document.createElement('canvas');
            return !!(window.WebGL2RenderingContext && canvas.getContext('webgl2'));
        }
        catch (e) {
            return false;
        }
    },
});

class NgtCommonLight extends NgtObject {
    set args(v) {
        this.instanceArgs = v;
    }
    set intensity(intensity) {
        this.set({ intensity: coerceNumberProperty(intensity) });
    }
    objectInitFn() {
        const { intensity, instanceArgs } = this.get();
        const light = new this.lightType(...instanceArgs);
        if (intensity != undefined) {
            light.intensity = intensity;
        }
        return light;
    }
    get preObjectInit() {
        return (initFn) => {
            this.set((state) => { var _a; return ({ intensity: (_a = state.intensity) !== null && _a !== void 0 ? _a : 1 }); });
            initFn();
        };
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { intensity: false });
    }
}
NgtCommonLight.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonLight, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonLight.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonLight, inputs: { args: "args", intensity: "intensity" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonLight, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }], intensity: [{
                type: Input
            }] } });

function provideCommonLightRef(subLightType, factory) {
    return [
        provideObjectRef(subLightType, factory),
        { provide: NgtCommonLight, useExisting: subLightType },
        {
            provide: NGT_COMMON_LIGHT_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subLightType],
        },
    ];
}

class NgtCommonLine extends NgtMaterialGeometry {
    get objectType() {
        return this.lineType;
    }
}
NgtCommonLine.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonLine, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonLine.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonLine, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonLine, decorators: [{
            type: Directive
        }] });

function provideMaterialGeometryObjectRef(subMaterialGeometryType, factory) {
    return [
        provideObjectRef(subMaterialGeometryType, factory),
        { provide: NgtMaterialGeometry, useExisting: subMaterialGeometryType },
        {
            provide: NGT_MATERIAL_GEOMETRY_OBJECT_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subMaterialGeometryType],
        },
    ];
}

function provideCommonLineRef(subLineType, factory) {
    return [
        provideMaterialGeometryObjectRef(subLineType, factory),
        { provide: NgtCommonLine, useExisting: subLineType },
        {
            provide: NGT_COMMON_LINE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subLineType],
        },
    ];
}

class NgtCommonMaterial extends NgtInstance {
    constructor(zone, store, parentRef, parentHostRef) {
        super(zone, store, parentRef, parentHostRef);
        this.init = this.effect(tapEffect(() => {
            const material = this.prepareInstance(new this.materialType());
            return () => {
                material.dispose();
            };
        }));
        this.setParameters = this.effect(tap(({ parameters, material }) => {
            material.setValues(Object.assign(Object.entries(parameters).reduce((result, [key, value]) => {
                if (value !== undefined) {
                    result[key] = value;
                }
                return result;
            }, {}), 'uniforms' in material && 'uniforms' in parameters
                ? {
                    uniforms: Object.assign(Object.assign({}, material.uniforms), parameters.uniforms),
                }
                : {}));
            material.needsUpdate = true;
        }));
        // TODO: remove when [parameters] is removed
        this.parameters$ = this.select(this.select((s) => s.alphaTest).pipe(startWithUndefined()), this.select((s) => s.alphaToCoverage).pipe(startWithUndefined()), this.select((s) => s.blendDst).pipe(startWithUndefined()), this.select((s) => s.blendDstAlpha).pipe(startWithUndefined()), this.select((s) => s.blendEquation).pipe(startWithUndefined()), this.select((s) => s.blendEquationAlpha).pipe(startWithUndefined()), this.select((s) => s.blending).pipe(startWithUndefined()), this.select((s) => s.blendSrc).pipe(startWithUndefined()), this.select((s) => s.blendSrcAlpha).pipe(startWithUndefined()), this.select((s) => s.clipIntersection).pipe(startWithUndefined()), this.select((s) => s.clippingPlanes).pipe(startWithUndefined()), this.select((s) => s.clipShadows).pipe(startWithUndefined()), this.select((s) => s.colorWrite).pipe(startWithUndefined()), this.select((s) => s.defines).pipe(startWithUndefined()), this.select((s) => s.depthFunc).pipe(startWithUndefined()), this.select((s) => s.depthTest).pipe(startWithUndefined()), this.select((s) => s.depthWrite).pipe(startWithUndefined()), this.select((s) => s.fog).pipe(startWithUndefined()), this.select((s) => s.name).pipe(startWithUndefined()), this.select((s) => s.opacity).pipe(startWithUndefined()), this.select((s) => s.polygonOffset).pipe(startWithUndefined()), this.select((s) => s.polygonOffsetFactor).pipe(startWithUndefined()), this.select((s) => s.polygonOffsetUnits).pipe(startWithUndefined()), this.select((s) => s.precision).pipe(startWithUndefined()), this.select((s) => s.premultipliedAlpha).pipe(startWithUndefined()), this.select((s) => s.dithering).pipe(startWithUndefined()), this.select((s) => s.side).pipe(startWithUndefined()), this.select((s) => s.shadowSide).pipe(startWithUndefined()), this.select((s) => s.toneMapped).pipe(startWithUndefined()), this.select((s) => s.transparent).pipe(startWithUndefined()), this.select((s) => s.vertexColors).pipe(startWithUndefined()), this.select((s) => s.visible).pipe(startWithUndefined()), this.select((s) => s.format).pipe(startWithUndefined()), this.select((s) => s.stencilWrite).pipe(startWithUndefined()), this.select((s) => s.stencilFunc).pipe(startWithUndefined()), this.select((s) => s.stencilRef).pipe(startWithUndefined()), this.select((s) => s.stencilWriteMask).pipe(startWithUndefined()), this.select((s) => s.stencilFuncMask).pipe(startWithUndefined()), this.select((s) => s.stencilFail).pipe(startWithUndefined()), this.select((s) => s.stencilZFail).pipe(startWithUndefined()), this.select((s) => s.stencilZPass).pipe(startWithUndefined()), this.select((s) => s.userData).pipe(startWithUndefined()), (alphaTest, alphaToCoverage, blendDst, blendDstAlpha, blendEquation, blendEquationAlpha, blending, blendSrc, blendSrcAlpha, clipIntersection, clippingPlanes, clipShadows, colorWrite, defines, depthFunc, depthTest, depthWrite, fog, name, opacity, polygonOffset, polygonOffsetFactor, polygonOffsetUnits, precision, premultipliedAlpha, dithering, side, shadowSide, toneMapped, transparent, vertexColors, visible, format, stencilWrite, stencilFunc, stencilRef, stencilWriteMask, stencilFuncMask, stencilFail, stencilZFail, stencilZPass, userData) => ({
            alphaTest,
            alphaToCoverage,
            blendDst,
            blendDstAlpha,
            blendEquation,
            blendEquationAlpha,
            blending,
            blendSrc,
            blendSrcAlpha,
            clipIntersection,
            clippingPlanes,
            clipShadows,
            colorWrite,
            defines,
            depthFunc,
            depthTest,
            depthWrite,
            fog,
            name,
            opacity,
            polygonOffset,
            polygonOffsetFactor,
            polygonOffsetUnits,
            precision,
            premultipliedAlpha,
            dithering,
            side,
            shadowSide,
            toneMapped,
            transparent,
            vertexColors,
            visible,
            format,
            stencilWrite,
            stencilFunc,
            stencilRef,
            stencilWriteMask,
            stencilFuncMask,
            stencilFail,
            stencilZFail,
            stencilZPass,
            userData,
        }));
    }
    set alphaTest(alphaTest) {
        this.set({ alphaTest: coerceNumberProperty(alphaTest) });
    }
    set alphaToCoverage(alphaToCoverage) {
        this.set({ alphaToCoverage: coerceBooleanProperty(alphaToCoverage) });
    }
    set blendDst(blendDst) {
        this.set({ blendDst });
    }
    set blendDstAlpha(blendDstAlpha) {
        this.set({ blendDstAlpha: coerceNumberProperty(blendDstAlpha) });
    }
    set blendEquation(blendEquation) {
        this.set({ blendEquation });
    }
    set blendEquationAlpha(blendEquationAlpha) {
        this.set({
            blendEquationAlpha: coerceNumberProperty(blendEquationAlpha),
        });
    }
    set blending(blending) {
        this.set({ blending });
    }
    set blendSrc(blendSrc) {
        this.set({ blendSrc });
    }
    set blendSrcAlpha(blendSrcAlpha) {
        this.set({ blendSrcAlpha: coerceNumberProperty(blendSrcAlpha) });
    }
    set clipIntersection(clipIntersection) {
        this.set({ clipIntersection: coerceBooleanProperty(clipIntersection) });
    }
    set clippingPlanes(clippingPlanes) {
        this.set({ clippingPlanes });
    }
    set clipShadows(clipShadows) {
        this.set({ clipShadows: coerceBooleanProperty(clipShadows) });
    }
    set colorWrite(colorWrite) {
        this.set({ colorWrite: coerceBooleanProperty(colorWrite) });
    }
    set defines(defines) {
        this.set({ defines });
    }
    set depthFunc(depthFunc) {
        this.set({ depthFunc });
    }
    set depthTest(depthTest) {
        this.set({ depthTest: coerceBooleanProperty(depthTest) });
    }
    set depthWrite(depthWrite) {
        this.set({ depthWrite: coerceBooleanProperty(depthWrite) });
    }
    set fog(fog) {
        this.set({ fog: coerceBooleanProperty(fog) });
    }
    set name(name) {
        this.set({ name });
    }
    set opacity(opacity) {
        this.set({ opacity: coerceNumberProperty(opacity) });
    }
    set polygonOffset(polygonOffset) {
        this.set({ polygonOffset: coerceBooleanProperty(polygonOffset) });
    }
    set polygonOffsetFactor(polygonOffsetFactor) {
        this.set({
            polygonOffsetFactor: coerceNumberProperty(polygonOffsetFactor),
        });
    }
    set polygonOffsetUnits(polygonOffsetUnits) {
        this.set({
            polygonOffsetUnits: coerceNumberProperty(polygonOffsetUnits),
        });
    }
    set precision(precision) {
        this.set({ precision });
    }
    set premultipliedAlpha(premultipliedAlpha) {
        this.set({
            premultipliedAlpha: coerceBooleanProperty(premultipliedAlpha),
        });
    }
    set dithering(dithering) {
        this.set({ dithering: coerceBooleanProperty(dithering) });
    }
    set side(side) {
        this.set({ side });
    }
    set shadowSide(shadowSide) {
        this.set({ shadowSide });
    }
    set toneMapped(toneMapped) {
        this.set({ toneMapped: coerceBooleanProperty(toneMapped) });
    }
    set transparent(transparent) {
        this.set({ transparent: coerceBooleanProperty(transparent) });
    }
    set vertexColors(vertexColors) {
        this.set({ vertexColors: coerceBooleanProperty(vertexColors) });
    }
    set visible(visible) {
        this.set({ visible: coerceBooleanProperty(visible) });
    }
    set format(format) {
        this.set({ format });
    }
    set stencilWrite(stencilWrite) {
        this.set({ stencilWrite: coerceBooleanProperty(stencilWrite) });
    }
    set stencilFunc(stencilFunc) {
        this.set({ stencilFunc });
    }
    set stencilRef(stencilRef) {
        this.set({ stencilRef: coerceNumberProperty(stencilRef) });
    }
    set stencilWriteMask(stencilWriteMask) {
        this.set({ stencilWriteMask: coerceNumberProperty(stencilWriteMask) });
    }
    set stencilFuncMask(stencilFuncMask) {
        this.set({ stencilFuncMask: coerceNumberProperty(stencilFuncMask) });
    }
    set stencilFail(stencilFail) {
        this.set({ stencilFail });
    }
    set stencilZFail(stencilZFail) {
        this.set({ stencilZFail });
    }
    set stencilZPass(stencilZPass) {
        this.set({ stencilZPass });
    }
    set userData(userData) {
        this.set({ userData });
    }
    /**
     * @deprecated Use individual inputs instead. Notice: Do not mix [parameters] and individual inputs, they will not be merged. Will be removed in next major version
     */
    set parameters(v) {
        this.set({ materialParameters: v });
    }
    get parameters() {
        return this.get((s) => s.materialParameters);
    }
    preInit() {
        this.set((state) => ({
            attach: state.attach.length ? state.attach : ['material'],
            materialParameters: state.materialParameters || {},
        }));
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                this.init();
                this.postInit();
                this.setParameters(this.select(this.instance$, this.select((s) => s.materialParameters), this.parameters$, this.subParameters$, (material, materialParameters, parameters, subParameters) => ({
                    material,
                    parameters: Object.keys(materialParameters).length
                        ? materialParameters
                        : Object.assign(Object.assign({}, parameters), subParameters),
                })));
            });
        });
    }
    // TODO: put these back after [parameters] is removed. Right now, Material needs to handle their own update options logic
    // protected override get optionFields(): Record<string, boolean> {
    //     return { ...super.optionFields,
    //         alphaTest: true,
    //         alphaToCoverage: true,
    //         blendDst: true,
    //         blendDstAlpha: true,
    //         blendEquation: true,
    //         blendEquationAlpha: true,
    //         blending: true,
    //         blendSrc: true,
    //         blendSrcAlpha: true,
    //         clipIntersection: true,
    //         clippingPlanes: true,
    //         clipShadows: true,
    //         colorWrite: true,
    //         defines: true,
    //         depthFunc: true,
    //         depthTest: true,
    //         depthWrite: true,
    //         fog: true,
    //         name: true,
    //         opacity: true,
    //         polygonOffset: true,
    //         polygonOffsetFactor: true,
    //         polygonOffsetUnits: true,
    //         precision: true,
    //         premultipliedAlpha: true,
    //         dithering: true,
    //         side: true,
    //         shadowSide: true,
    //         toneMapped: true,
    //         transparent: true,
    //         vertexColors: true,
    //         visible: true,
    //         format: true,
    //         stencilWrite: true,
    //         stencilFunc: true,
    //         stencilRef: true,
    //         stencilWriteMask: true,
    //         stencilFuncMask: true,
    //         stencilFail: true,
    //         stencilZFail: true,
    //         stencilZPass: true,
    //         userData: true,
    //     };
    // }
    // TODO: remove when [parameters] is removed
    get subParameters$() {
        const optionFieldEntries = Object.entries(this.optionFields);
        if (optionFieldEntries.length === 0)
            return of({});
        return this.select(...optionFieldEntries.map(([inputKey, shouldStartWithUndefined]) => {
            const option$ = this.select((s) => s[inputKey]);
            if (shouldStartWithUndefined)
                return option$.pipe(startWithUndefined());
            return option$;
        }), (...args) => args.reduce((record, arg, index) => {
            record[optionFieldEntries[index][0]] = arg;
            return record;
        }, {}));
    }
}
NgtCommonMaterial.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonMaterial, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_OBJECT_REF, optional: true, skipSelf: true }, { token: NGT_OBJECT_HOST_REF, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonMaterial.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonMaterial, inputs: { alphaTest: "alphaTest", alphaToCoverage: "alphaToCoverage", blendDst: "blendDst", blendDstAlpha: "blendDstAlpha", blendEquation: "blendEquation", blendEquationAlpha: "blendEquationAlpha", blending: "blending", blendSrc: "blendSrc", blendSrcAlpha: "blendSrcAlpha", clipIntersection: "clipIntersection", clippingPlanes: "clippingPlanes", clipShadows: "clipShadows", colorWrite: "colorWrite", defines: "defines", depthFunc: "depthFunc", depthTest: "depthTest", depthWrite: "depthWrite", fog: "fog", name: "name", opacity: "opacity", polygonOffset: "polygonOffset", polygonOffsetFactor: "polygonOffsetFactor", polygonOffsetUnits: "polygonOffsetUnits", precision: "precision", premultipliedAlpha: "premultipliedAlpha", dithering: "dithering", side: "side", shadowSide: "shadowSide", toneMapped: "toneMapped", transparent: "transparent", vertexColors: "vertexColors", visible: "visible", format: "format", stencilWrite: "stencilWrite", stencilFunc: "stencilFunc", stencilRef: "stencilRef", stencilWriteMask: "stencilWriteMask", stencilFuncMask: "stencilFuncMask", stencilFail: "stencilFail", stencilZFail: "stencilZFail", stencilZPass: "stencilZPass", userData: "userData", parameters: "parameters" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonMaterial, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_HOST_REF]
                    }] }];
    }, propDecorators: { alphaTest: [{
                type: Input
            }], alphaToCoverage: [{
                type: Input
            }], blendDst: [{
                type: Input
            }], blendDstAlpha: [{
                type: Input
            }], blendEquation: [{
                type: Input
            }], blendEquationAlpha: [{
                type: Input
            }], blending: [{
                type: Input
            }], blendSrc: [{
                type: Input
            }], blendSrcAlpha: [{
                type: Input
            }], clipIntersection: [{
                type: Input
            }], clippingPlanes: [{
                type: Input
            }], clipShadows: [{
                type: Input
            }], colorWrite: [{
                type: Input
            }], defines: [{
                type: Input
            }], depthFunc: [{
                type: Input
            }], depthTest: [{
                type: Input
            }], depthWrite: [{
                type: Input
            }], fog: [{
                type: Input
            }], name: [{
                type: Input
            }], opacity: [{
                type: Input
            }], polygonOffset: [{
                type: Input
            }], polygonOffsetFactor: [{
                type: Input
            }], polygonOffsetUnits: [{
                type: Input
            }], precision: [{
                type: Input
            }], premultipliedAlpha: [{
                type: Input
            }], dithering: [{
                type: Input
            }], side: [{
                type: Input
            }], shadowSide: [{
                type: Input
            }], toneMapped: [{
                type: Input
            }], transparent: [{
                type: Input
            }], vertexColors: [{
                type: Input
            }], visible: [{
                type: Input
            }], format: [{
                type: Input
            }], stencilWrite: [{
                type: Input
            }], stencilFunc: [{
                type: Input
            }], stencilRef: [{
                type: Input
            }], stencilWriteMask: [{
                type: Input
            }], stencilFuncMask: [{
                type: Input
            }], stencilFail: [{
                type: Input
            }], stencilZFail: [{
                type: Input
            }], stencilZPass: [{
                type: Input
            }], userData: [{
                type: Input
            }], parameters: [{
                type: Input
            }] } });

function provideCommonMaterialRef(subMaterialType, factory) {
    return [
        provideInstanceRef(subMaterialType, factory),
        { provide: NgtCommonMaterial, useExisting: subMaterialType },
        {
            provide: NGT_COMMON_MATERIAL_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subMaterialType],
        },
    ];
}

class NgtCommonMesh extends NgtMaterialGeometry {
    get objectType() {
        return this.meshType;
    }
}
NgtCommonMesh.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonMesh, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonMesh.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonMesh, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonMesh, decorators: [{
            type: Directive
        }] });

function provideCommonMeshRef(subMeshType, factory) {
    return [
        provideMaterialGeometryObjectRef(subMeshType, factory),
        { provide: NgtCommonMesh, useExisting: subMeshType },
        {
            provide: NGT_COMMON_MESH_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subMeshType],
        },
    ];
}

class NgtCommonObjectHelper extends NgtInstance {
    constructor(zone, store, parentRef, parentHostRef) {
        super(zone, store, parentRef, parentHostRef);
        this.init = this.effect(tapEffect((instanceArgs) => {
            if (!this.parent || !this.parent.value) {
                console.info('Parent is not an object3d');
                return;
            }
            const objectHelper = this.prepareInstance(new this.objectHelperType(this.parent.value, ...instanceArgs));
            const scene = this.store.get((s) => s.scene);
            if (objectHelper && scene) {
                scene.add(objectHelper);
                const unregister = this.store.registerBeforeRender({
                    callback: () => {
                        if (objectHelper) {
                            objectHelper.update();
                        }
                    },
                    obj: this.instance,
                });
                return () => {
                    if (objectHelper && scene) {
                        scene.remove(objectHelper);
                        unregister();
                    }
                };
            }
            return;
        }));
    }
    set helperVisible(helperVisible) {
        this.set({ visible: coerceBooleanProperty(helperVisible) });
    }
    preInit() {
        super.preInit();
        this.set((state) => {
            var _a;
            return ({
                visible: (_a = state['visible']) !== null && _a !== void 0 ? _a : true,
            });
        });
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                this.init(this.instanceArgs$);
                this.postInit();
            });
        });
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { visible: false });
    }
}
NgtCommonObjectHelper.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonObjectHelper, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_OBJECT_REF, optional: true }, { token: NGT_OBJECT_HOST_REF, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonObjectHelper.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonObjectHelper, inputs: { helperVisible: "helperVisible" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonObjectHelper, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_OBJECT_HOST_REF]
                    }] }];
    }, propDecorators: { helperVisible: [{
                type: Input
            }] } });

function provideCommonObjectHelperRef(subObjectHelperType, factory) {
    return [
        provideInstanceRef(subObjectHelperType, factory),
        { provide: NgtCommonObjectHelper, useExisting: subObjectHelperType },
        {
            provide: NGT_COMMON_OBJECT_HELPER_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subObjectHelperType],
        },
    ];
}

function provideSkinnedMeshHostRef(subType, factory) {
    return {
        provide: NGT_HOST_SKINNED_MESH_REF,
        useFactory: (instance) => {
            return () => factory(instance);
        },
        deps: [subType],
    };
}
function provideSkeletonHostRef(subType, factory) {
    return {
        provide: NGT_HOST_SKELETON_REF,
        useFactory: (instance) => {
            return () => factory(instance);
        },
        deps: [subType],
    };
}
function provideBoneHostRef(subType, factory) {
    return {
        provide: NGT_HOST_BONE_REF,
        useFactory: (instance) => {
            return () => factory(instance);
        },
        deps: [subType],
    };
}

class NgtCommonSprite extends NgtObject {
    set args(v) {
        this.instanceArgs = v;
    }
    set material(material) {
        this.set({ material });
    }
    objectInitFn() {
        const { material } = this.get();
        return new this.spriteType(material);
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { material: true });
    }
}
NgtCommonSprite.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonSprite, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonSprite.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonSprite, inputs: { args: "args", material: "material" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonSprite, decorators: [{
            type: Directive
        }], propDecorators: { args: [{
                type: Input
            }], material: [{
                type: Input
            }] } });

function provideCommonSpriteRef(subSpriteType, factory) {
    return [
        provideObjectRef(subSpriteType, factory),
        { provide: NgtCommonSprite, useExisting: subSpriteType },
        {
            provide: NGT_COMMON_SPRITE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subSpriteType],
        },
    ];
}

class NgtCommonTexture extends NgtInstance {
    constructor(zone, store, parentRef, parentHostRef, document) {
        super(zone, store, parentRef, parentHostRef);
        this.document = document;
        this.init = this.effect(tapEffect(() => {
            const gl = this.store.get((s) => s.gl);
            const instanceArgs = this.get((s) => s.instanceArgs);
            const textureInstanceArgs = this.adjustCtorParams(instanceArgs);
            const texture = this.prepareInstance(new this.textureType(...textureInstanceArgs));
            texture.encoding = gl.outputEncoding;
            texture.needsUpdate = true;
            return () => {
                texture.dispose();
            };
        }));
    }
    set args(v) {
        this.instanceArgs = v;
    }
    set image(image) {
        this.set({ image });
    }
    set mapping(mapping) {
        this.set({ mapping });
    }
    set wrapS(wrapS) {
        this.set({ wrapS });
    }
    set wrapT(wrapT) {
        this.set({ wrapT });
    }
    set magFilter(magFilter) {
        this.set({ magFilter });
    }
    set minFilter(minFilter) {
        this.set({ minFilter });
    }
    set format(format) {
        this.set({ format });
    }
    set type(type) {
        this.set({ type });
    }
    set anisotropy(anisotropy) {
        this.set({ anisotropy: coerceNumberProperty(anisotropy) });
    }
    set encoding(encoding) {
        this.set({ encoding });
    }
    ngOnInit() {
        super.ngOnInit();
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.store.ready$, () => {
                this.init(this.ctorParams$);
                this.postInit();
            });
        });
    }
    get optionFields() {
        return Object.assign(Object.assign({}, super.optionFields), { image: true, mapping: true, wrapS: true, wrapT: true, magFilter: true, minFilter: true, format: true, type: true, anisotropy: true, encoding: true });
    }
}
NgtCommonTexture.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonTexture, deps: [{ token: i0.NgZone }, { token: NgtStore }, { token: NGT_INSTANCE_REF, optional: true, skipSelf: true }, { token: NGT_INSTANCE_HOST_REF, optional: true, skipSelf: true }, { token: DOCUMENT }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonTexture.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCommonTexture, inputs: { args: "args", image: "image", mapping: "mapping", wrapS: "wrapS", wrapT: "wrapT", magFilter: "magFilter", minFilter: "minFilter", format: "format", type: "type", anisotropy: "anisotropy", encoding: "encoding" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCommonTexture, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_REF]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_INSTANCE_HOST_REF]
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }];
    }, propDecorators: { args: [{
                type: Input
            }], image: [{
                type: Input
            }], mapping: [{
                type: Input
            }], wrapS: [{
                type: Input
            }], wrapT: [{
                type: Input
            }], magFilter: [{
                type: Input
            }], minFilter: [{
                type: Input
            }], format: [{
                type: Input
            }], type: [{
                type: Input
            }], anisotropy: [{
                type: Input
            }], encoding: [{
                type: Input
            }] } });

function provideCommonTextureRef(subTextureType, factory) {
    return [
        provideInstanceRef(subTextureType, factory),
        { provide: NgtCommonTexture, useExisting: subTextureType },
        {
            provide: NGT_COMMON_TEXTURE_REF,
            useFactory: (instance) => {
                return () => (factory === null || factory === void 0 ? void 0 : factory(instance)) || instance.instance;
            },
            deps: [subTextureType],
        },
    ];
}

class NgtCursor {
    constructor(object, document) {
        if (!object)
            return;
        merge(object.pointerover.pipe(map(() => true)), object.pointerout.pipe(map(() => false)))
            .pipe(takeUntil(object.destroy$))
            .subscribe((hovered) => {
            document.body.style.cursor = hovered ? 'pointer' : 'auto';
        });
    }
}
NgtCursor.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursor, deps: [{ token: NgtObject, optional: true, self: true }, { token: DOCUMENT }], target: i0.ɵɵFactoryTarget.Directive });
NgtCursor.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtCursor, selector: "[ngtCursor]", exportAs: ["ngtCursor"], ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursor, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngtCursor]',
                    exportAs: 'ngtCursor',
                }]
        }], ctorParameters: function () {
        return [{ type: NgtObject, decorators: [{
                        type: Optional
                    }, {
                        type: Self
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }];
    } });
class NgtCursorModule {
}
NgtCursorModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursorModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCursorModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursorModule, declarations: [NgtCursor], exports: [NgtCursor] });
NgtCursorModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursorModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtCursorModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtCursor],
                    exports: [NgtCursor],
                }]
        }] });

class NgtRepeat extends NgForOf {
    set ngForRepeat(count) {
        this.ngForOf = Number.isInteger(count)
            ? Array.from({ length: count }, (_, i) => i)
            : [];
    }
}
NgtRepeat.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeat, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtRepeat.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtRepeat, selector: "[ngFor][ngForRepeat]", inputs: { ngForRepeat: "ngForRepeat" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeat, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngFor][ngForRepeat]',
                }]
        }], propDecorators: { ngForRepeat: [{
                type: Input
            }] } });
class NgtRepeatModule {
}
NgtRepeatModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeatModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtRepeatModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeatModule, declarations: [NgtRepeat], exports: [NgtRepeat] });
NgtRepeatModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeatModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRepeatModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtRepeat],
                    exports: [NgtRepeat],
                }]
        }] });

class NgtObjectPassThrough {
    constructor(host) {
        this.host = host;
        if (!host)
            return;
    }
    set ngtObjectOutputs(wrapper) {
        this.assertWrapper(wrapper);
        if (wrapper.click.observed) {
            this.host.click
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.click.emit.bind(wrapper.click));
        }
        if (wrapper.contextmenu.observed) {
            this.host.contextmenu
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.contextmenu.emit.bind(wrapper.contextmenu));
        }
        if (wrapper.dblclick.observed) {
            this.host.dblclick
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.dblclick.emit.bind(wrapper.dblclick));
        }
        if (wrapper.pointerup.observed) {
            this.host.pointerup
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerup.emit.bind(wrapper.pointerup));
        }
        if (wrapper.pointerdown.observed) {
            this.host.pointerdown
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerdown.emit.bind(wrapper.pointerdown));
        }
        if (wrapper.pointerover.observed) {
            this.host.pointerover
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerover.emit.bind(wrapper.pointerover));
        }
        if (wrapper.pointerout.observed) {
            this.host.pointerout
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerout.emit.bind(wrapper.pointerout));
        }
        if (wrapper.pointerenter.observed) {
            this.host.pointerenter
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerenter.emit.bind(wrapper.pointerenter));
        }
        if (wrapper.pointerleave.observed) {
            this.host.pointerleave
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointerleave.emit.bind(wrapper.pointerleave));
        }
        if (wrapper.pointermove.observed) {
            this.host.pointermove
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointermove.emit.bind(wrapper.pointermove));
        }
        if (wrapper.pointermissed.observed) {
            this.host.pointermissed
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointermissed.emit.bind(wrapper.pointermissed));
        }
        if (wrapper.pointercancel.observed) {
            this.host.pointercancel
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.pointercancel.emit.bind(wrapper.pointercancel));
        }
        if (wrapper.wheel.observed) {
            this.host.wheel
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.wheel.emit.bind(wrapper.wheel));
        }
        if (wrapper.ready.observed) {
            this.host.ready
                .pipe(takeUntil(wrapper.destroy$))
                .subscribe(wrapper.ready.emit.bind(wrapper.ready));
        }
    }
    set ngtObjectInputs(wrapper) {
        this.assertWrapper(wrapper);
        if (wrapper.shouldPassThroughRef) {
            this.host.ref = wrapper.instance;
        }
        merge(wrapper.select((s) => s.attach), this.host.select((s) => s.attach))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.attach = this.host.get((s) => s.attachExplicit)
                ? this.host.attach
                : wrapper.attach;
        });
        merge(wrapper.select((s) => s.skipParent), this.host.select((s) => s.skipParent))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.skipParent = this.host.get((s) => s.skipParentExplicit)
                ? this.host.skipParent
                : wrapper.skipParent;
        });
        merge(wrapper.select((s) => s.noAttach), this.host.select((s) => s.noAttach))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.noAttach = this.host.get((s) => s.noAttachExplicit)
                ? this.host.noAttach
                : wrapper.noAttach;
        });
        merge(wrapper.select((s) => s.name), this.host.select((s) => s.name))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.name = this.host.get((s) => s.nameExplicit)
                ? this.host.name
                : wrapper.name;
        });
        merge(wrapper.select((s) => s.position), this.host.select((s) => s.position))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.position = this.host.get((s) => s.positionExplicit)
                ? this.host.position
                : wrapper.position;
        });
        merge(wrapper.select((s) => s.rotation), this.host.select((s) => s.rotation))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.rotation = this.host.get((s) => s.rotationExplicit)
                ? this.host.rotation
                : wrapper.rotation;
        });
        merge(wrapper.select((s) => s.quaternion), this.host.select((s) => s.quaternion))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.quaternion = this.host.get((s) => s.quaternionExplicit)
                ? this.host.quaternion
                : wrapper.quaternion;
        });
        merge(wrapper.select((s) => s.scale), this.host.select((s) => s.scale))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.scale = this.host.get((s) => s.scaleExplicit)
                ? this.host.scale
                : wrapper.scale;
        });
        merge(wrapper.select((s) => s.color), this.host.select((s) => s.color))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.color = this.host.get((s) => s.colorExplicit)
                ? this.host.color
                : wrapper.color;
        });
        merge(wrapper.select((s) => s.userData), this.host.select((s) => s.userData))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.userData = this.host.get((s) => s.userDataExplicit)
                ? this.host.userData
                : wrapper.userData;
        });
        merge(wrapper.select((s) => s.castShadow).pipe(startWithUndefined()), this.host.select((s) => s.castShadow).pipe(startWithUndefined()))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.castShadow = this.host.get((s) => s.castShadowExplicit)
                ? this.host.castShadow
                : wrapper.castShadow;
        });
        merge(wrapper.select((s) => s.receiveShadow).pipe(startWithUndefined()), this.host.select((s) => s.receiveShadow).pipe(startWithUndefined()))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.receiveShadow = this.host.get((s) => s.receiveShadowExplicit)
                ? this.host.receiveShadow
                : wrapper.receiveShadow;
        });
        merge(wrapper.select((s) => s.visible), this.host.select((s) => s.visible))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.visible = this.host.get((s) => s.visibleExplicit)
                ? this.host.visible
                : wrapper.visible;
        });
        merge(wrapper.select((s) => s.matrixAutoUpdate), this.host.select((s) => s.matrixAutoUpdate))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.matrixAutoUpdate = this.host.get((s) => s.matrixAutoUpdateExplicit)
                ? this.host.matrixAutoUpdate
                : wrapper.matrixAutoUpdate;
        });
        merge(wrapper.select((s) => s.dispose).pipe(startWithUndefined()), this.host.select((s) => s.dispose).pipe(startWithUndefined()))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.dispose = this.host.get((s) => s.disposeExplicit)
                ? this.host.dispose
                : wrapper.dispose;
        });
        merge(wrapper.select((s) => s.raycast).pipe(startWithUndefined()), this.host.select((s) => s.raycast).pipe(startWithUndefined()))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.raycast = this.host.get((s) => s.raycastExplicit)
                ? this.host.raycast
                : wrapper.raycast;
        });
        merge(wrapper.select((s) => s.appendMode), this.host.select((s) => s.appendMode))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.appendMode = this.host.get((s) => s.appendModeExplicit)
                ? this.host.appendMode
                : wrapper.appendMode;
        });
        merge(wrapper.select((s) => s.appendTo).pipe(startWithUndefined()), this.host.select((s) => s.appendTo).pipe(startWithUndefined()))
            .pipe(takeUntil(wrapper.destroy$))
            .subscribe(() => {
            this.host.appendTo = this.host.get((s) => s.appendToExplicit)
                ? this.host.appendTo
                : wrapper.appendTo;
        });
    }
    assertWrapper(wrapper) {
        if (!(wrapper instanceof NgtObjectInputs)) {
            throw new Error(`[NgtObjectPassThrough] wrapper is not an NgtObject`);
        }
    }
}
NgtObjectPassThrough.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThrough, deps: [{ token: NgtObject, optional: true, self: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtObjectPassThrough.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.3.4", type: NgtObjectPassThrough, selector: "[ngtObjectOutputs],[ngtObjectInputs],[ngtObjectInputs][ngtObjectOutputs]", inputs: { ngtObjectOutputs: "ngtObjectOutputs", ngtObjectInputs: "ngtObjectInputs" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThrough, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngtObjectOutputs],[ngtObjectInputs],[ngtObjectInputs][ngtObjectOutputs]',
                }]
        }], ctorParameters: function () {
        return [{ type: NgtObject, decorators: [{
                        type: Optional
                    }, {
                        type: Self
                    }] }];
    }, propDecorators: { ngtObjectOutputs: [{
                type: Input
            }], ngtObjectInputs: [{
                type: Input
            }] } });
class NgtObjectPassThroughModule {
}
NgtObjectPassThroughModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThroughModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtObjectPassThroughModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThroughModule, declarations: [NgtObjectPassThrough], exports: [NgtObjectPassThrough] });
NgtObjectPassThroughModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThroughModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtObjectPassThroughModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtObjectPassThrough],
                    exports: [NgtObjectPassThrough],
                }]
        }] });

/**
 * @deprecated Use {@link NgtColorAttribute} instead
 */
class NgtColorPipe {
    /**
     * ConstructorParameters<typeof THREE.Color> has a limitation on THREE.Color constructor overloads
     * @deprecated
     */
    transform(args) {
        return makeColor(args);
    }
}
NgtColorPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtColorPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipe, name: "color" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'color' }]
        }] });
class NgtColorPipeModule {
}
NgtColorPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtColorPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipeModule, declarations: [NgtColorPipe], exports: [NgtColorPipe] });
NgtColorPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtColorPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtColorPipe],
                    exports: [NgtColorPipe],
                }]
        }] });

/**
 * @deprecated Use {@link NgtFogAttribute} instead
 */
class NgtFogPipe {
    /**
     * @deprecated
     * @param value
     */
    transform(value) {
        return new THREE.Fog(...value);
    }
}
NgtFogPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtFogPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipe, name: "fog" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'fog',
                }]
        }] });
/**
 * @deprecated Use {@link NgtFogExp2Attribute} instead
 */
class NgtFogExp2Pipe {
    /**
     * @deprecated
     * @param value
     */
    transform(value) {
        return new THREE.FogExp2(...value);
    }
}
NgtFogExp2Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogExp2Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtFogExp2Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogExp2Pipe, name: "fogExp2" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogExp2Pipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'fogExp2',
                }]
        }] });
class NgtFogPipeModule {
}
NgtFogPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtFogPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipeModule, declarations: [NgtFogPipe, NgtFogExp2Pipe], exports: [NgtFogPipe, NgtFogExp2Pipe] });
NgtFogPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtFogPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtFogPipe, NgtFogExp2Pipe],
                    exports: [NgtFogPipe, NgtFogExp2Pipe],
                }]
        }] });

class NgtMathPipe {
    transform(value, keyOfMath) {
        const params = is.arr(value) ? value : [value];
        return Math[keyOfMath](...params);
    }
}
NgtMathPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtMathPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipe, name: "math" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'math' }]
        }] });
class NgtMathPipeModule {
}
NgtMathPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtMathPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipeModule, declarations: [NgtMathPipe], exports: [NgtMathPipe] });
NgtMathPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtMathPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtMathPipe],
                    exports: [NgtMathPipe],
                }]
        }] });

class NgtPiPipe {
    transform(value) {
        return value * Math.PI;
    }
}
NgtPiPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtPiPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipe, name: "pi" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'pi' }]
        }] });
class NgtPiPipeModule {
}
NgtPiPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtPiPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipeModule, declarations: [NgtPiPipe], exports: [NgtPiPipe] });
NgtPiPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtPiPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtPiPipe],
                    exports: [NgtPiPipe],
                }]
        }] });

class NgtRadianPipe {
    transform(degree) {
        return (degree * Math.PI) / 180;
    }
}
NgtRadianPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtRadianPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipe, name: "radian" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'radian' }]
        }] });
class NgtRadianPipeModule {
}
NgtRadianPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtRadianPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipeModule, declarations: [NgtRadianPipe], exports: [NgtRadianPipe] });
NgtRadianPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtRadianPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtRadianPipe],
                    exports: [NgtRadianPipe],
                }]
        }] });

class NgtSidePipe {
    transform(side) {
        switch (side) {
            case 'front':
                return THREE.FrontSide;
            case 'back':
                return THREE.BackSide;
            case 'double':
                return THREE.DoubleSide;
        }
    }
}
NgtSidePipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtSidePipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipe, name: "side" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipe, decorators: [{
            type: Pipe,
            args: [{ name: 'side' }]
        }] });
class NgtSidePipeModule {
}
NgtSidePipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtSidePipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipeModule, declarations: [NgtSidePipe], exports: [NgtSidePipe] });
NgtSidePipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtSidePipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtSidePipe],
                    exports: [NgtSidePipe],
                }]
        }] });

/**
 * @deprecated Use {@link NgtVector2Attribute} instead
 */
class NgtVector2Pipe {
    /**
     * @deprecated
     * @param value
     */
    transform(value) {
        return makeVector2(value);
    }
}
NgtVector2Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector2Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector2Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector2Pipe, name: "vector2" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector2Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector2' }]
        }] });
/**
 * @deprecated Use {@link NgtVector3Attribute} instead
 */
class NgtVector3Pipe {
    /**
     * @deprecated
     * @param value
     */
    transform(value) {
        return makeVector3(value);
    }
}
NgtVector3Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector3Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector3Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector3Pipe, name: "vector3" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector3Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector3' }]
        }] });
/**
 * @deprecated Use {@link NgtVector4Attribute} instead
 */
class NgtVector4Pipe {
    /**
     * @deprecated
     * @param value
     */
    transform(value) {
        return makeVector4(value);
    }
}
NgtVector4Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector4Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector4Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector4Pipe, name: "vector4" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVector4Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector4' }]
        }] });
class NgtVectorPipeModule {
}
NgtVectorPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVectorPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtVectorPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVectorPipeModule, declarations: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe], exports: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe] });
NgtVectorPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVectorPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtVectorPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe],
                    exports: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe],
                }]
        }] });

function buildGraph(object) {
    const data = { nodes: {}, materials: {} };
    if (object) {
        object.traverse((obj) => {
            if (obj['name']) {
                data.nodes[obj['name']] = obj;
            }
            if (obj['material'] && !data.materials[obj['material']['name']]) {
                data.materials[obj['material']['name']] = obj['material'];
            }
        });
    }
    return data;
}

class NgtLoader {
    constructor() {
        this.cached = new Map();
    }
    use(loaderConstructor, input, extensions, onProgress) {
        const keys = (is.arr(input) ? input : [input]);
        const loader = new loaderConstructor();
        if (extensions) {
            extensions(loader);
        }
        const results$ = forkJoin(keys.map((key) => {
            if (this.cached.has(key)) {
                return of(this.cached.get(key));
            }
            return defer(() => loader.loadAsync(key, onProgress)).pipe(tap((data) => {
                if (data.scene) {
                    Object.assign(data, buildGraph(data.scene));
                }
                this.cached.set(key, data);
            }), catchError((err) => {
                console.error(`Error loading ${key}: ${err.message}`);
                return throwError(err);
            }));
        }));
        return defer(() => is.arr(input)
            ? results$
            : results$.pipe(map((results) => results[0]))).pipe(share({
            connector: () => new ReplaySubject(),
            resetOnRefCountZero: true,
            resetOnError: true,
        }));
    }
    ngOnDestroy() {
        this.cached.clear();
    }
}
NgtLoader.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtLoader, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgtLoader.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtLoader, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.4", ngImport: i0, type: NgtLoader, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }] });

function shaderMaterial(uniforms, vertexShader, fragmentShader, onInit) {
    return class extends THREE.ShaderMaterial {
        constructor() {
            const entries = Object.entries(uniforms);
            // Create unforms and shaders
            super({
                uniforms: entries.reduce((acc, [name, value]) => {
                    const uniform = THREE.UniformsUtils.clone({
                        [name]: { value },
                    });
                    return Object.assign(Object.assign({}, acc), uniform);
                }, {}),
                vertexShader,
                fragmentShader,
            });
            // Create getter/setters
            entries.forEach(([name]) => Object.defineProperty(this, name, {
                get: () => this.uniforms[name].value,
                set: (v) => (this.uniforms[name].value = v),
            }));
            if (onInit)
                onInit(this);
        }
    };
}

/* abstracts */

/**
 * Generated bundle index. Do not edit.
 */

export { NGT_CAMERA_REF, NGT_COMMON_ATTRIBUTE_REF, NGT_COMMON_AUDIO_REF, NGT_COMMON_CAMERA_REF, NGT_COMMON_CURVE_REF, NGT_COMMON_GEOMETRY_REF, NGT_COMMON_HELPER_REF, NGT_COMMON_LIGHT_REF, NGT_COMMON_LINE_REF, NGT_COMMON_MATERIAL_REF, NGT_COMMON_MESH_REF, NGT_COMMON_OBJECT_HELPER_REF, NGT_COMMON_SPRITE_REF, NGT_COMMON_TEXTURE_REF, NGT_HOST_BONE_REF, NGT_HOST_SKELETON_REF, NGT_HOST_SKINNED_MESH_REF, NGT_INSTANCE_HOST_REF, NGT_INSTANCE_REF, NGT_IS_WEBGL_AVAILABLE, NGT_MATERIAL_GEOMETRY_OBJECT_REF, NGT_OBJECT_HOST_REF, NGT_OBJECT_REF, NGT_PERFORMANCE_OPTIONS, NGT_RESIZE_BOX_DEFAULT, NGT_RESIZE_DEBOUNCE_DEFAULT, NGT_RESIZE_OBSERVER_SUPPORT, NGT_RESIZE_OPTIONS, NGT_SCENE_REF, NgtCanvas, NgtCanvasModule, NgtColorPipe, NgtColorPipeModule, NgtCommonAttribute, NgtCommonAudio, NgtCommonCamera, NgtCommonCurve, NgtCommonGeometry, NgtCommonHelper, NgtCommonLight, NgtCommonLine, NgtCommonMaterial, NgtCommonMesh, NgtCommonObjectHelper, NgtCommonSprite, NgtCommonTexture, NgtComponentStore, NgtCoreModule, NgtCursor, NgtCursorModule, NgtFogExp2Pipe, NgtFogPipe, NgtFogPipeModule, NgtInstance, NgtLoader, NgtMaterialGeometry, NgtMathPipe, NgtMathPipeModule, NgtObject, NgtObjectInputs, NgtObjectPassThrough, NgtObjectPassThroughModule, NgtPiPipe, NgtPiPipeModule, NgtPortal, NgtPortalContent, NgtPortalModule, NgtRadianPipe, NgtRadianPipeModule, NgtRepeat, NgtRepeatModule, NgtResize, NgtSidePipe, NgtSidePipeModule, NgtStore, NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe, NgtVectorPipeModule, Ref, WINDOW, addAfterCallback, addCallback, addTailCallback, applyProps, buildGraph, coerceBooleanProperty, coerceNumberProperty, createDefaultCamera, createLoop, debounceSync, getInstanceRootState, is, make, makeColor, makeDpr, makeId, makeVector2, makeVector3, makeVector4, prepare, privateKeys, provideBoneHostRef, provideCanvasInstanceRef, provideCommonAttributeRef, provideCommonAudioRef, provideCommonCameraRef, provideCommonCurveRef, provideCommonGeometryRef, provideCommonHelperRef, provideCommonLightRef, provideCommonLineRef, provideCommonMaterialRef, provideCommonMeshRef, provideCommonObjectHelperRef, provideCommonSpriteRef, provideCommonTextureRef, provideInstanceHostRef, provideInstanceRef, provideMaterialGeometryObjectRef, provideObjectHostRef, provideObjectRef, providePerformanceOptions, provideResizeOptions, provideSkeletonHostRef, provideSkinnedMeshHostRef, render, shaderMaterial, skipUndefined, startWithUndefined, tapEffect, updateCamera };
//# sourceMappingURL=angular-three-core.mjs.map
